import numpy as np
from ..utils import check_supported


class PaddingBase:
    """
    A class for performing padding based on coordinate transform.
    The Cuda and OpenCL backends will subclass this class.
    """

    supported_modes = ["constant", "edge", "reflect", "symmetric", "wrap"]

    def __init__(self, shape, pad_width, mode="constant", **kwargs):
        """
        Initialize a Padding object.

        Parameters
        ----------
        shape: tuple
            Image shape
        pad_width: tuple
            Padding width for each axis. Please see the documentation of numpy.pad().
        mode: str
            Padding mode

        Other parameters
        ----------------
        constant_values: tuple
            Tuple containing the values to fill when mode="constant" (as in numpy.pad)
        """
        if len(shape) != 2:
            raise ValueError("This class only works on images")
        self.shape = shape
        self._set_mode(mode, **kwargs)
        self._get_padding_arrays(pad_width)

    def _set_mode(self, mode, **kwargs):
        # COMPAT.
        if mode == "edges":
            mode = "edge"
        #
        check_supported(mode, self.supported_modes, "padding mode")
        self.mode = mode
        self._kwargs = kwargs

    def _get_padding_arrays(self, pad_width):
        self.pad_width = pad_width
        if isinstance(pad_width, tuple) and isinstance(pad_width[0], np.ndarray):
            # user-defined coordinate transform
            err_msg = "pad_width must be either a scalar, a tuple in the form ((a, b), (c, d)), or a tuple of two one-dimensional numpy arrays (eg. use numpy.indices(..., sparse=True))"
            if len(pad_width) != 2:
                raise ValueError(err_msg)
            if any([np.squeeze(pw).ndim > 1 for pw in pad_width]):
                raise ValueError(err_msg)
            if self.mode == "constant":
                raise ValueError("Custom coordinate transform does not work with mode='constant'")
            self.mode = "custom"
            self.coords_rows, self.coords_cols = pad_width
        else:
            if self.mode == "constant":
                # no need for coordinate transform here
                constant_values = self._kwargs.get("constant_values", 0)
                self.padded_array_constant = np.pad(
                    np.zeros(self.shape, dtype="f"), self.pad_width, mode="constant", constant_values=constant_values
                )
                self.padded_shape = self.padded_array_constant.shape
                return
            R, C = np.indices(self.shape, dtype=np.int32, sparse=True)
            self.coords_rows = np.pad(R.ravel(), self.pad_width[0], mode=self.mode)
            self.coords_cols = np.pad(C.ravel(), self.pad_width[1], mode=self.mode)
        self.padded_shape = (self.coords_rows.size, self.coords_cols.size)



import numpy as np
import pytest
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.opencl.utils import __has_pyopencl__, get_opencl_context
from nabu.testutils import get_data, generate_tests_scenarios, __do_long_tests__
from nabu.processing.roll_opencl import OpenCLRoll

configs_roll = {
    "shape": [(300, 451), (300, 300), (255, 300)],
    "offset_x": [0, 10, 155],
    "dtype": [np.float32],  # , np.complex64],
}


scenarios_roll = generate_tests_scenarios(configs_roll)


@pytest.fixture(scope="class")
def bootstrap_roll(request):
    cls = request.cls
    cls.data = get_data("chelsea.npz")["data"]
    cls.tol = 1e-7
    if __has_pycuda__:
        cls.cu_ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.cl_ctx = get_opencl_context(device_type="all")
    yield
    if __has_pycuda__:
        cls.cu_ctx.pop()


@pytest.mark.usefixtures("bootstrap_roll")
class TestRoll:
    @staticmethod
    def _compute_ref(data, direction, offset):
        ref = data.copy()
        ref[:, offset:] = np.roll(data[:, offset:], direction, axis=1)
        return ref

    @pytest.mark.skipif(not (__has_pyopencl__), reason="Need pyopencl for this test")
    @pytest.mark.parametrize("config", scenarios_roll)
    def test_opencl_roll(self, config):
        shape = config["shape"]
        dtype = config["dtype"]
        offset_x = config["offset_x"]
        data = np.ascontiguousarray(self.data[: shape[0], : shape[1]], dtype=dtype)

        ref_forward = self._compute_ref(data, 1, offset_x)
        ref_backward = self._compute_ref(data, -1, offset_x)

        roll_forward = OpenCLRoll(dtype, direction=1, offset=offset_x, ctx=self.cl_ctx)
        d_data = roll_forward.processing.allocate_array("data", data.shape, dtype=dtype)
        d_data.set(data)
        roll_backward = OpenCLRoll(dtype, direction=-1, offset=offset_x, queue=roll_forward.processing.queue)

        roll_forward(d_data)
        # from spire.utils import ims
        # ims([d_data.get(), ref_forward, d_data.get() - ref_forward])
        assert np.allclose(d_data.get(), ref_forward), "roll_forward: something wrong with config=%s" % (str(config))

        d_data.set(data)
        roll_backward(d_data)
        assert np.allclose(d_data.get(), ref_backward), "roll_backward: something wrong with config=%s" % (str(config))



from os import path
from ..resources.logger import LoggerOrPrint
from .utils import get_subregion
from .writer import WriterManager
from ..io.reader import get_hdf5_dataset_shape

try:
    import pycuda.gpuarray as garray

    __has_pycuda__ = True
except:
    __has_pycuda__ = False


class DataDumpManager:
    """
    A helper class for managing data dumps, with the aim of saving/resuming the processing from a given step.
    """

    def __init__(self, process_config, sub_region, margin=None, logger=None):
        """
        Initialize a DataDump object.

        Parameters
        -----------
        process_config: ProcessConfig
            ProcessConfig object
        sub_region: tuple of int
            Series of integers defining the sub-region being processed.
            The form is ((start_angle, end_angle), (start_z, end_z), (start_x, end_x))
        margin: tuple of int, optional
            Margin, used when processing data, in the form ((up, down), (left, right)).
            Each item can be None.
            Using a margin means that a given chunk of data will eventually be cropped as
            `data[:, up:-down, left:-right]`
        logger: Logger, optional
            Logging object
        """
        self.process_config = process_config
        self.processing_steps = process_config.processing_steps
        self.processing_options = process_config.processing_options
        self.dataset_info = process_config.dataset_info

        self._set_subregion_and_margin(sub_region, margin)
        self.logger = LoggerOrPrint(logger)
        self._configure_data_dumps()

    def _set_subregion_and_margin(self, sub_region, margin):
        self.sub_region = get_subregion(sub_region)
        self._z_sub_region = self.sub_region[1]
        self.z_min = self._z_sub_region[0]
        self.margin = get_subregion(margin, ndim=2)  # ((U, D), (L, R))
        self.margin_up = self.margin[0][0] or 0
        self.start_index = self.z_min + self.margin_up
        self.delta_z = self._z_sub_region[-1] - self._z_sub_region[-2]

        self._grouped_processing = False
        iangle1, iangle2 = self.sub_region[0]
        if iangle1 != 0 or iangle2 < len(self.process_config.rotation_angles(subsampling=False)):
            self._grouped_processing = True
            self.start_index = self.sub_region[0][0]

    def _configure_dump(self, step_name, force_dump_to_fname=None):
        if force_dump_to_fname is not None:
            # Shortcut
            fname_full = force_dump_to_fname
        elif step_name in self.processing_steps:
            # Standard case
            if not self.processing_options[step_name].get("save", False):
                return
            fname_full = self.processing_options[step_name]["save_steps_file"]
        elif step_name == "sinogram" and self.process_config.dump_sinogram:
            # "sinogram" is a special keyword
            fname_full = self.process_config.dump_sinogram_file
        else:
            return

        # "fname_full" is the path to the final master file.
        # We also need to create partial files (in a sub-directory)
        fname, ext = path.splitext(fname_full)
        dirname, file_prefix = path.split(fname)

        self.data_dump[step_name] = WriterManager(
            dirname,
            file_prefix,
            file_format="hdf5",
            overwrite=True,
            start_index=self.start_index,
            logger=self.logger,
            metadata={
                "process_name": step_name,
                "processing_index": 0,
                "config": {
                    "processing_options": self.processing_options,  # slow!
                    "nabu_config": self.process_config.nabu_config,
                },
                "entry": getattr(self.dataset_info.dataset_scanner, "entry", "entry"),
            },
        )

    def _configure_data_dumps(self):
        self.data_dump = {}
        for step_name in self.processing_steps:
            self._configure_dump(step_name)
        # sinogram is a special keyword: not in processing_steps, but guaranteed to be before sinogram generation
        if self.process_config.dump_sinogram:
            self._configure_dump("sinogram")

    def get_data_dump(self, step_name):
        """
        Get information on where to write a given processing step.

        Parameters
        ----------
        step_name: str
            Name of the processing step

        Returns
        -------
        writer_configurator: WriterConfigurator
            An object with information on where to write the data for the given processing step.
        """
        return self.data_dump.get(step_name, None)

    def get_read_dump_subregion(self):
        read_opts = self.processing_options["read_chunk"]
        if read_opts.get("process_file", None) is None:
            return None
        dump_start_z, dump_end_z = read_opts["dump_start_z"], read_opts["dump_end_z"]
        relative_start_z = self.z_min - dump_start_z
        relative_end_z = relative_start_z + self.delta_z
        # When using binning, every step after "read" results in smaller-sized data.
        # Therefore dumped data has shape (ceil(n_angles/subsampling), n_z//binning_z, n_x//binning_x)
        relative_start_z //= self.process_config.binning_z
        relative_end_z //= self.process_config.binning_z
        # (n_angles, n_z, n_x)
        subregion = (None, None, relative_start_z, relative_end_z, None, None)
        return subregion

    def _check_resume_from_step(self):
        read_opts = self.processing_options["read_chunk"]
        expected_radios_shape = get_hdf5_dataset_shape(
            read_opts["process_file"],
            read_opts["process_h5_path"],
            sub_region=self.get_read_dump_subregion(),
        )
        # TODO check

    def dump_data_to_file(self, step_name, data, crop_margin=False):
        if step_name not in self.data_dump:
            return
        writer = self.data_dump[step_name]
        self.logger.info("Dumping data to %s" % writer.fname)
        if __has_pycuda__:
            if isinstance(data, garray.GPUArray):
                data = data.get()

        margin_up = self.margin[0][0] or None
        margin_down = self.margin[0][1] or None
        margin_down = -margin_down if margin_down is not None else None  # pylint: disable=E1130
        if crop_margin and (margin_up is not None or margin_down is not None):
            data = data[:, margin_up:margin_down, :]
        metadata = {"dump_sub_region": {"sub_region": self.sub_region, "margin": self.margin}}
        writer.write_data(data, metadata=metadata)

    def __repr__(self):
        res = "%s(%s, margin=%s)" % (self.__class__.__name__, str(self.sub_region), str(self.margin))
        if len(self.data_dump) > 0:
            for step_name, writer_configurator in self.data_dump.items():
                res += "\n- Dump %s to %s" % (step_name, writer_configurator.fname)
        return res



import numpy
import logging
import h5py
import os
from typing import Iterable
from silx.io.url import DataUrl
from silx.io.utils import get_data
from datetime import datetime

from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxtransformations import NXtransformations
from nxtomo.utils.transformation import build_matrix, DetYFlipTransformation, DetZFlipTransformation
from nxtomo.paths.nxtomo import get_paths as _get_nexus_paths

from tomoscan.io import HDF5File
from tomoscan.series import Series
from tomoscan.esrf import NXtomoScan, EDFTomoScan
from tomoscan.esrf.scan.utils import (
    get_compacted_dataslices,
)  # this version has a 'return_url_set' needed here. At one point they should be merged together
from nabu.stitching.config import (
    PreProcessedSingleAxisStitchingConfiguration,
    KEY_IMG_REG_METHOD,
)
from nabu.stitching.utils import find_projections_relative_shifts
from functools import lru_cache as cache
from .single_axis import SingleAxisStitcher
from pyunitsystem.metricsystem import MetricSystem


_logger = logging.getLogger(__name__)


class PreProcessingStitching(SingleAxisStitcher):
    """
    loader to be used when save data during pre-processing stitching (on projections). Output is expected to be an NXtomo

    warning: axis are provided according to the `acquisition space <https://tomo.gitlab-pages.esrf.fr/bliss-tomo/master/modelization.html>`_
    """

    def __init__(self, configuration, progress=None) -> None:
        """ """
        if not isinstance(configuration, PreProcessedSingleAxisStitchingConfiguration):
            raise TypeError(
                f"configuration is expected to be an instance of {PreProcessedSingleAxisStitchingConfiguration}. Get {type(configuration)} instead"
            )
        super().__init__(configuration, progress=progress)
        self._series = Series("series", iterable=configuration.input_scans, use_identifiers=False)
        self._reading_orders = []
        # TODO: rename flips to axis_0_flips, axis_1_flips, axis_2_flips...
        self._x_flips = []
        self._y_flips = []
        self._z_flips = []

        # 'expend' auto shift request if only set once for all
        if numpy.isscalar(self.configuration.axis_0_pos_px):
            self.configuration.axis_0_pos_px = [
                self.configuration.axis_0_pos_px,
            ] * (len(self.series) - 1)
        if numpy.isscalar(self.configuration.axis_1_pos_px):
            self.configuration.axis_1_pos_px = [
                self.configuration.axis_1_pos_px,
            ] * (len(self.series) - 1)
        if numpy.isscalar(self.configuration.axis_1_pos_px):
            self.configuration.axis_1_pos_px = [
                self.configuration.axis_1_pos_px,
            ] * (len(self.series) - 1)

        if self.configuration.axis_0_params is None:
            self.configuration.axis_0_params = {}
        if self.configuration.axis_1_params is None:
            self.configuration.axis_1_params = {}
        if self.configuration.axis_2_params is None:
            self.configuration.axis_2_params = {}

    def pre_processing_computation(self):
        self.compute_reduced_flats_and_darks()

    @property
    def stitching_axis_in_frame_space(self):
        if self.axis == 0:
            return 0
        elif self.axis == 1:
            return 1
        elif self.axis == 2:
            raise NotImplementedError(
                "pre-processing stitching along axis 2 is not handled. This would require to do interpolation between frame along the rotation angle. Just not possible"
            )
        else:
            raise NotImplementedError(f"stitching axis must be in (0, 1, 2). Get {self.axis}")

    @property
    def x_flips(self) -> list:
        return self._x_flips

    @property
    def y_flips(self) -> list:
        return self._y_flips

    def order_input_tomo_objects(self):

        def get_min_bound(scan):
            return scan.get_bounding_box(axis=self.axis).min

        # order scans along the stitched axis
        if self.axis == 0:
            position_along_stitched_axis = self.configuration.axis_0_pos_px
        elif self.axis == 1:
            position_along_stitched_axis = self.configuration.axis_1_pos_px
        else:
            raise ValueError(
                "stitching cannot be done along axis 2 for pre-processing. This would require to interpolate frame between different rotation angle"
            )
        # if axis 0 position is provided then use directly it
        if position_along_stitched_axis is not None and len(position_along_stitched_axis) > 0:
            order = numpy.argsort(position_along_stitched_axis)[::-1]
            sorted_series = Series(
                self.series.name,
                numpy.take_along_axis(numpy.array(self.series[:]), order, axis=0),
                use_identifiers=False,
            )
        else:
            # else use bounding box
            sorted_series = Series(
                self.series.name,
                sorted(self.series[:], key=get_min_bound, reverse=True),
                use_identifiers=False,
            )
        if sorted_series != self.series:
            if sorted_series[:] != self.series[::-1]:
                raise ValueError(
                    f"Unable to get comprehensive input. Axis {self.axis} (decreasing) ordering is not respected."
                )
            else:
                _logger.warning(
                    f"decreasing order haven't been respected. Need to reorder {self.serie_label} ({[str(scan) for scan in sorted_series[:]]}). Will also reorder overlap height, stitching height and invert shifts"
                )
                if self.configuration.axis_0_pos_mm is not None:
                    self.configuration.axis_0_pos_mm = self.configuration.axis_0_pos_mm[::-1]
                if self.configuration.axis_0_pos_px is not None:
                    self.configuration.axis_0_pos_px = self.configuration.axis_0_pos_px[::-1]
                if self.configuration.axis_1_pos_mm is not None:
                    self.configuration.axis_1_pos_mm = self.configuration.axis_1_pos_mm[::-1]
                if self.configuration.axis_1_pos_px is not None:
                    self.configuration.axis_1_pos_px = self.configuration.axis_1_pos_px[::-1]
                if self.configuration.axis_2_pos_mm is not None:
                    self.configuration.axis_2_pos_mm = self.configuration.axis_2_pos_mm[::-1]
                if self.configuration.axis_2_pos_px is not None:
                    self.configuration.axis_2_pos_px = self.configuration.axis_2_pos_px[::-1]
                if not numpy.isscalar(self._configuration.flip_ud):
                    self._configuration.flip_ud = self._configuration.flip_ud[::-1]
                if not numpy.isscalar(self._configuration.flip_lr):
                    self._configuration.flip_ud = self._configuration.flip_lr[::-1]

        self._series = sorted_series

    def check_inputs(self):
        """
        insure input data is coherent
        """
        n_scans = len(self.series)
        if n_scans == 0:
            raise ValueError("no scan to stich together")

        for scan in self.series:
            from tomoscan.scanbase import TomoScanBase

            if not isinstance(scan, TomoScanBase):
                raise TypeError(f"z-preproc stitching expects instances of {TomoScanBase}. {type(scan)} provided.")

        # check output file path and data path are provided
        if self.configuration.output_file_path in (None, ""):
            raise ValueError("output_file_path should be provided to the configuration")
        if self.configuration.output_data_path in (None, ""):
            raise ValueError("output_data_path should be provided to the configuration")

        # check number of shift provided
        for axis_pos_px, axis_name in zip(
            (
                self.configuration.axis_0_pos_px,
                self.configuration.axis_1_pos_px,
                self.configuration.axis_1_pos_px,
                self.configuration.axis_0_pos_mm,
                self.configuration.axis_1_pos_mm,
                self.configuration.axis_2_pos_mm,
            ),
            (
                "axis_0_pos_px",
                "axis_1_pos_px",
                "axis_2_pos_px",
                "axis_0_pos_mm",
                "axis_1_pos_mm",
                "axis_2_pos_mm",
            ),
        ):
            if isinstance(axis_pos_px, Iterable) and len(axis_pos_px) != (n_scans):
                raise ValueError(f"{axis_name} expect {n_scans} shift defined. Get {len(axis_pos_px)}")

        self._reading_orders = []
        # the first scan will define the expected reading orderd, and expected flip.
        # if all scan are flipped then we will keep it this way
        self._reading_orders.append(1)

        # check scans are coherent (nb projections, rotation angle, energy...)
        for scan_0, scan_1 in zip(self.series[0:-1], self.series[1:]):
            if len(scan_0.projections) != len(scan_1.projections):
                raise ValueError(f"{scan_0} and {scan_1} have a different number of projections")
            if isinstance(scan_0, NXtomoScan) and isinstance(scan_1, NXtomoScan):
                # check rotation (only of is an NXtomoScan)
                scan_0_angles = numpy.asarray(scan_0.rotation_angle)
                scan_0_projections_angles = scan_0_angles[
                    numpy.asarray(scan_0.image_key_control) == ImageKey.PROJECTION.value
                ]
                scan_1_angles = numpy.asarray(scan_1.rotation_angle)
                scan_1_projections_angles = scan_1_angles[
                    numpy.asarray(scan_1.image_key_control) == ImageKey.PROJECTION.value
                ]
                if not numpy.allclose(scan_0_projections_angles, scan_1_projections_angles, atol=10e-1):
                    if numpy.allclose(
                        scan_0_projections_angles,
                        scan_1_projections_angles[::-1],
                        atol=10e-1,
                    ):
                        reading_order = -1 * self._reading_orders[-1]
                    else:
                        raise ValueError(f"Angles from {scan_0} and {scan_1} are different")
                else:
                    reading_order = 1 * self._reading_orders[-1]
                self._reading_orders.append(reading_order)
            # check energy
            if scan_0.energy is None:
                _logger.warning(f"no energy found for {scan_0}")
            elif not numpy.isclose(scan_0.energy, scan_1.energy, rtol=1e-03):
                _logger.warning(
                    f"different energy found between {scan_0} ({scan_0.energy}) and {scan_1} ({scan_1.energy})"
                )
            # check FOV
            if not scan_0.field_of_view == scan_1.field_of_view:
                raise ValueError(f"{scan_0} and {scan_1} have different field of view")
            # check distance
            if scan_0.distance is None:
                _logger.warning(f"no distance found for {scan_0}")
            elif not numpy.isclose(scan_0.distance, scan_1.distance, rtol=10e-3):
                raise ValueError(f"{scan_0} and {scan_1} have different sample / detector distance")
            # check pixel size
            if not numpy.isclose(scan_0.x_pixel_size, scan_1.x_pixel_size):
                raise ValueError(
                    f"{scan_0} and {scan_1} have different x pixel size. {scan_0.x_pixel_size} vs {scan_1.x_pixel_size}"
                )
            if not numpy.isclose(scan_0.y_pixel_size, scan_1.y_pixel_size):
                raise ValueError(
                    f"{scan_0} and {scan_1} have different y pixel size. {scan_0.y_pixel_size} vs {scan_1.y_pixel_size}"
                )

        for scan in self.series:
            # check x, y and z translation are constant (only if is an NXtomoScan)
            if isinstance(scan, NXtomoScan):
                if scan.x_translation is not None and not numpy.isclose(
                    min(scan.x_translation), max(scan.x_translation)
                ):
                    _logger.warning(
                        "x translations appears to be evolving over time. Might end up with wrong stitching"
                    )
                if scan.y_translation is not None and not numpy.isclose(
                    min(scan.y_translation), max(scan.y_translation)
                ):
                    _logger.warning(
                        "y translations appears to be evolving over time. Might end up with wrong stitching"
                    )
                if scan.z_translation is not None and not numpy.isclose(
                    min(scan.z_translation), max(scan.z_translation)
                ):
                    _logger.warning(
                        "z translations appears to be evolving over time. Might end up with wrong stitching"
                    )

    def _compute_positions_as_px(self):
        """insure we have or we can deduce an estimated position as pixel"""

        def get_position_as_px_on_axis(axis, pos_as_px, pos_as_mm):
            if pos_as_px is not None:
                if pos_as_mm is not None:
                    raise ValueError(
                        f"position of axis {axis} is provided twice: as mm and as px. Please provide one only ({pos_as_mm} vs {pos_as_px})"
                    )
                else:
                    return pos_as_px

            elif pos_as_mm is not None:
                # deduce from position given in configuration and pixel size
                axis_N_pos_px = []
                for scan, pos_in_mm in zip(self.series, pos_as_mm):
                    pixel_size_m = self.configuration.pixel_size or scan.pixel_size
                    axis_N_pos_px.append((pos_in_mm / MetricSystem.MILLIMETER.value) / pixel_size_m)
                return axis_N_pos_px
            else:
                # deduce from motor position and pixel size
                axis_N_pos_px = []
                base_position_m = self.series[0].get_bounding_box(axis=axis).min
                for scan in self.series:
                    pixel_size_m = self.configuration.pixel_size or scan.pixel_size
                    scan_axis_bb = scan.get_bounding_box(axis=axis)
                    axis_N_mean_pos_m = (scan_axis_bb.max - scan_axis_bb.min) / 2 + scan_axis_bb.min
                    axis_N_mean_rel_pos_m = axis_N_mean_pos_m - base_position_m
                    axis_N_pos_px.append(int(axis_N_mean_rel_pos_m / pixel_size_m))
                return axis_N_pos_px

        for axis, property_px_name, property_mm_name in zip(
            (0, 1, 2),
            (
                "axis_0_pos_px",
                "axis_1_pos_px",
                "axis_2_pos_px",
            ),
            (
                "axis_0_pos_mm",
                "axis_1_pos_mm",
                "axis_2_pos_mm",
            ),
        ):
            assert hasattr(
                self.configuration, property_px_name
            ), f"configuration API changed. should have {property_px_name}"
            assert hasattr(
                self.configuration, property_mm_name
            ), f"configuration API changed. should have {property_px_name}"
            try:
                new_px_position = get_position_as_px_on_axis(
                    axis=axis,
                    pos_as_px=getattr(self.configuration, property_px_name),
                    pos_as_mm=getattr(self.configuration, property_mm_name),
                )
            except ValueError:
                # when unable to find the position
                if axis == self.axis:
                    # if we cannot find position over the stitching axis then raise an error: unable to process without
                    raise
                else:
                    _logger.warning(f"Unable to find position over axis {axis}. Set them to zero")
                    setattr(
                        self.configuration,
                        property_px_name,
                        numpy.array([0] * len(self.series)),
                    )
            else:
                setattr(
                    self.configuration,
                    property_px_name,
                    new_px_position,
                )

        # clear position in mm as the one we will used are the px one
        self.configuration.axis_0_pos_mm = None
        self.configuration.axis_1_pos_mm = None
        self.configuration.axis_2_pos_mm = None

        # add some log
        if self.configuration.axis_2_pos_mm is not None or self.configuration.axis_2_pos_px is not None:
            _logger.warning("axis 2 position is not handled by the stitcher. Will be ignored")
        axis_0_pos = ", ".join([f"{pos}px" for pos in self.configuration.axis_0_pos_px])
        axis_1_pos = ", ".join([f"{pos}px" for pos in self.configuration.axis_1_pos_px])
        axis_2_pos = ", ".join([f"{pos}px" for pos in self.configuration.axis_2_pos_px])
        _logger.info(f"axis 0 position to be used: " + axis_0_pos)
        _logger.info(f"axis 1 position to be used: " + axis_1_pos)
        _logger.info(f"axis 2 position to be used: " + axis_2_pos)
        _logger.info(f"stitching will be applied along axis: {self.axis}")

    def compute_estimated_shifts(self):
        if self.axis == 0:
            # if we want to stitch over axis 0 (aka z)
            axis_0_pos_px = self.configuration.axis_0_pos_px
            self._axis_0_rel_ini_shifts = []
            # compute overlap along axis 0
            for upper_scan, lower_scan, upper_scan_axis_0_pos, lower_scan_axis_0_pos in zip(
                self.series[:-1], self.series[1:], axis_0_pos_px[:-1], axis_0_pos_px[1:]
            ):
                upper_scan_pos = upper_scan_axis_0_pos - upper_scan.dim_2 / 2
                lower_scan_high_pos = lower_scan_axis_0_pos + lower_scan.dim_2 / 2
                # simple test of overlap. More complete test are run by check_overlaps later
                if lower_scan_high_pos <= upper_scan_pos:
                    raise ValueError(f"no overlap found between {upper_scan} and {lower_scan}")
                self._axis_0_rel_ini_shifts.append(
                    int(lower_scan_high_pos - upper_scan_pos)  # overlap are expected to be int for now
                )
            self._axis_1_rel_ini_shifts = self.from_abs_pos_to_rel_pos(self.configuration.axis_1_pos_px)
            self._axis_2_rel_ini_shifts = [0.0] * (len(self.series) - 1)
        elif self.axis == 1:
            # if we want to stitch over axis 1 (aka Y in acquisition reference - which is x in frame reference)
            axis_1_pos_px = self.configuration.axis_1_pos_px
            self._axis_1_rel_ini_shifts = []
            # compute overlap along axis 0
            for left_scan, right_scan, left_scan_axis_1_pos, right_scan_axis_1_pos in zip(
                self.series[:-1], self.series[1:], axis_1_pos_px[:-1], axis_1_pos_px[1:]
            ):
                left_scan_pos = left_scan_axis_1_pos - left_scan.dim_1 / 2
                right_scan_high_pos = right_scan_axis_1_pos + right_scan.dim_1 / 2
                # simple test of overlap. More complete test are run by check_overlaps later
                if right_scan_high_pos <= left_scan_pos:
                    raise ValueError(f"no overlap found between {left_scan} and {right_scan}")
                self._axis_1_rel_ini_shifts.append(
                    int(right_scan_high_pos - left_scan_pos)  # overlap are expected to be int for now
                )
            self._axis_0_rel_ini_shifts = self.from_abs_pos_to_rel_pos(self.configuration.axis_0_pos_px)
            self._axis_2_rel_ini_shifts = [0.0] * (len(self.series) - 1)
        else:
            raise NotImplementedError("stitching only forseen for axis 0 and 1 for now")

    def _compute_shifts(self):
        """
        compute all shift requested (set to 'auto' in the configuration)

        """
        n_scans = len(self.configuration.input_scans)
        if n_scans == 0:
            raise ValueError("no scan to stich provided")

        projection_for_shift = self.configuration.slice_for_cross_correlation or "middle"
        if self.axis not in (0, 1):
            raise NotImplementedError("only stitching over axis 0 and 2 are handled for pre-processing stitching")

        final_rel_shifts = []
        for (
            scan_0,
            scan_1,
            order_s0,
            order_s1,
            x_rel_shift,
            y_rel_shift,
        ) in zip(
            self.series[:-1],
            self.series[1:],
            self.reading_orders[:-1],
            self.reading_orders[1:],
            self._axis_1_rel_ini_shifts,
            self._axis_0_rel_ini_shifts,
        ):
            x_cross_algo = self.configuration.axis_1_params.get(KEY_IMG_REG_METHOD, None)
            y_cross_algo = self.configuration.axis_0_params.get(KEY_IMG_REG_METHOD, None)

            # compute relative shift
            found_shift_y, found_shift_x = find_projections_relative_shifts(
                upper_scan=scan_0,
                lower_scan=scan_1,
                projection_for_shift=projection_for_shift,
                x_cross_correlation_function=x_cross_algo,
                y_cross_correlation_function=y_cross_algo,
                x_shifts_params=self.configuration.axis_1_params,  # image x map acquisition axis 1 (Y)
                y_shifts_params=self.configuration.axis_0_params,  # image y map acquisition axis 0 (Z)
                invert_order=order_s1 != order_s0,
                estimated_shifts=(y_rel_shift, x_rel_shift),
                axis=self.axis,
            )
            final_rel_shifts.append(
                (found_shift_y, found_shift_x),
            )

        # set back values. Now position should start at 0
        self._axis_0_rel_final_shifts = [final_shift[0] for final_shift in final_rel_shifts]
        self._axis_1_rel_final_shifts = [final_shift[1] for final_shift in final_rel_shifts]
        self._axis_2_rel_final_shifts = [0.0] * len(final_rel_shifts)
        _logger.info(f"axis 1 relative shifts (x in radio ref) to be used will be {self._axis_0_rel_final_shifts}")
        print(f"axis 1 relative shifts (x in radio ref) to be used will be {self._axis_0_rel_final_shifts}")
        _logger.info(f"axis 0 relative shifts (y in radio ref) y to be used will be {self._axis_1_rel_final_shifts}")
        print(f"axis 0 relative shifts (y in radio ref) y to be used will be {self._axis_1_rel_final_shifts}")

    def _create_nx_tomo(self, store_composition: bool = False):
        """
        create final NXtomo with stitched frames.
        Policy: save all projections flat fielded. So this NXtomo will only contain projections (no dark and no flat).
        But nabu will be able to reconstruct it with field `flatfield` set to False
        """
        nx_tomo = NXtomo()

        nx_tomo.energy = self.series[0].energy
        start_times = list(filter(None, [scan.start_time for scan in self.series]))
        end_times = list(filter(None, [scan.end_time for scan in self.series]))

        if len(start_times) > 0:
            nx_tomo.start_time = (
                numpy.asarray([numpy.datetime64(start_time) for start_time in start_times]).min().astype(datetime)
            )
        else:
            _logger.warning("Unable to find any start_time from input")
        if len(end_times) > 0:
            nx_tomo.end_time = (
                numpy.asarray([numpy.datetime64(end_time) for end_time in end_times]).max().astype(datetime)
            )
        else:
            _logger.warning("Unable to find any end_time from input")

        title = ";".join([scan.sequence_name or "" for scan in self.series])
        nx_tomo.title = f"stitch done from {title}"

        self._slices_to_stitch, n_proj = self.configuration.settle_slices()

        # handle detector (without frames)
        nx_tomo.instrument.detector.field_of_view = self.series[0].field_of_view
        nx_tomo.instrument.detector.distance = self.series[0].distance
        nx_tomo.instrument.detector.x_pixel_size = self.series[0].x_pixel_size
        nx_tomo.instrument.detector.y_pixel_size = self.series[0].y_pixel_size
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_proj
        nx_tomo.instrument.detector.tomo_n = n_proj
        # note: stitching process insure un-flipping of frames. So make sure transformations is defined as an empty set
        nx_tomo.instrument.detector.transformations = NXtransformations()

        if isinstance(self.series[0], NXtomoScan):
            # note: first scan is always the reference as order to read data (so no rotation_angle inversion here)
            rotation_angle = numpy.asarray(self.series[0].rotation_angle)
            nx_tomo.sample.rotation_angle = rotation_angle[
                numpy.asarray(self.series[0].image_key_control) == ImageKey.PROJECTION.value
            ]
        elif isinstance(self.series[0], EDFTomoScan):
            nx_tomo.sample.rotation_angle = numpy.linspace(
                start=0, stop=self.series[0].scan_range, num=self.series[0].tomo_n
            )
        else:
            raise NotImplementedError(
                f"scan type ({type(self.series[0])} is not handled)",
                NXtomoScan,
                isinstance(self.series[0], NXtomoScan),
            )

        # do a sub selection of the rotation angle if a we are only computing a part of the slices
        def apply_slices_selection(array, slices, allow_empty: bool = False):
            if isinstance(slices, slice):
                return array[slices.start : slices.stop : 1]
            elif isinstance(slices, Iterable):
                return list([array[index] for index in slices])
            else:
                raise RuntimeError("slices must be instance of a slice or of an iterable")

        nx_tomo.sample.rotation_angle = apply_slices_selection(
            array=nx_tomo.sample.rotation_angle, slices=self._slices_to_stitch
        )

        # handle sample
        if False not in [isinstance(scan, NXtomoScan) for scan in self.series]:

            def get_sample_translation_for_projs(scan: NXtomoScan, attr):
                values = numpy.array(getattr(scan, attr))
                mask = scan.image_key_control == ImageKey.PROJECTION.value
                return values[mask]

            # we consider the new x, y and z position to be at the center of the one created
            x_translation = [
                get_sample_translation_for_projs(scan, "x_translation")
                for scan in self.series
                if scan.x_translation is not None
            ]
            if len(x_translation) > 0:
                # if there is some metadata about {x|y|z} translations
                # we want to take the mean of each frame for each projections
                x_translation = apply_slices_selection(
                    numpy.array(x_translation).mean(axis=0),
                    slices=self._slices_to_stitch,
                )
            else:
                # if no NXtomo has information about x_translation.
                # note: if at least one has missing values the numpy.Array(x_translation) with create an error as well
                x_translation = [0.0] * n_proj
                _logger.warning("Unable to fin input nxtomo x_translation values. Set it to 0.0")
            nx_tomo.sample.x_translation = x_translation

            y_translation = [
                get_sample_translation_for_projs(scan, "y_translation")
                for scan in self.series
                if scan.y_translation is not None
            ]
            if len(y_translation) > 0:
                y_translation = apply_slices_selection(
                    numpy.array(y_translation).mean(axis=0),
                    slices=self._slices_to_stitch,
                )
            else:
                y_translation = [0.0] * n_proj
                _logger.warning("Unable to fin input nxtomo y_translation values. Set it to 0.0")
            nx_tomo.sample.y_translation = y_translation
            z_translation = [
                get_sample_translation_for_projs(scan, "z_translation")
                for scan in self.series
                if scan.z_translation is not None
            ]
            if len(z_translation) > 0:
                z_translation = apply_slices_selection(
                    numpy.array(z_translation).mean(axis=0),
                    slices=self._slices_to_stitch,
                )
            else:
                z_translation = [0.0] * n_proj
                _logger.warning("Unable to fin input nxtomo z_translation values. Set it to 0.0")
            nx_tomo.sample.z_translation = z_translation

            nx_tomo.sample.name = self.series[0].sample_name

        # compute stitched frame shape
        if self.axis == 0:
            stitched_frame_shape = (
                n_proj,
                (
                    numpy.asarray([scan.dim_2 for scan in self.series]).sum()
                    - numpy.asarray([abs(overlap) for overlap in self._axis_0_rel_final_shifts]).sum()
                ),
                self._stitching_constant_length,
            )
        elif self.axis == 1:
            stitched_frame_shape = (
                n_proj,
                self._stitching_constant_length,
                (
                    numpy.asarray([scan.dim_1 for scan in self.series]).sum()
                    - numpy.asarray([abs(overlap) for overlap in self._axis_1_rel_final_shifts]).sum()
                ),
            )
        else:
            raise NotImplementedError("stitching on pre-processing along axis 2 (x-ray direction) is not handled")

        if stitched_frame_shape[0] < 1 or stitched_frame_shape[1] < 1 or stitched_frame_shape[2] < 1:
            raise RuntimeError(f"Error in stitched frame shape calculation. {stitched_frame_shape} found.")
        # get expected output dataset first (just in case output and input files are the same)
        first_proj_idx = sorted(self.series[0].projections.keys())[0]
        first_proj_url = self.series[0].projections[first_proj_idx]
        if h5py.is_hdf5(first_proj_url.file_path()):
            first_proj_url = DataUrl(
                file_path=first_proj_url.file_path(),
                data_path=first_proj_url.data_path(),
                scheme="h5py",
            )

        # first save the NXtomo entry without the frame
        # dicttonx will fail if the folder does not exists
        dir_name = os.path.dirname(self.configuration.output_file_path)
        if dir_name not in (None, ""):
            os.makedirs(dir_name, exist_ok=True)
        nx_tomo.save(
            file_path=self.configuration.output_file_path,
            data_path=self.configuration.output_data_path,
            nexus_path_version=self.configuration.output_nexus_version,
            overwrite=self.configuration.overwrite_results,
        )

        transformation_matrices = {
            scan.get_identifier()
            .to_str()
            .center(80, "-"): numpy.array2string(build_matrix(scan.get_detector_transformations(tuple())))
            for scan in self.series
        }
        _logger.info(
            "scan detector transformation matrices are:\n"
            "\n".join(["/n".join(item) for item in transformation_matrices.items()])
        )

        _logger.info(
            f"reading order is {self.reading_orders}",
        )

        def get_output_data_type():
            return numpy.float32  # because we will apply flat field correction on it and they are not raw data

        output_dtype = get_output_data_type()
        # append frames ("instrument/detector/data" dataset)
        with HDF5File(self.configuration.output_file_path, mode="a") as h5f:
            # note: nx_tomo.save already handles the possible overwrite conflict by removing
            # self.configuration.output_file_path or raising an error

            stitched_frame_path = "/".join(
                [
                    self.configuration.output_data_path,
                    _get_nexus_paths(self.configuration.output_nexus_version).PROJ_PATH,
                ]
            )
            self.dumper.output_dataset = h5f.create_dataset(
                name=stitched_frame_path,
                shape=stitched_frame_shape,
                dtype=output_dtype,
            )
            # TODO: we could also create in several time and create a virtual dataset from it.
            scans_projections_indexes = []
            for scan, reverse in zip(self.series, self.reading_orders):
                scans_projections_indexes.append(sorted(scan.projections.keys(), reverse=(reverse == -1)))
            if self.progress:
                self.progress.total = self.get_n_slices_to_stitch()

            if isinstance(self._slices_to_stitch, slice):
                step = self._slices_to_stitch.step or 1
            else:
                step = 1
            i_proj = 0
            for bunch_start, bunch_end in self._data_bunch_iterator(slices=self._slices_to_stitch, bunch_size=50):
                for data_frames in self._get_bunch_of_data(
                    bunch_start,
                    bunch_end,
                    step=step,
                    scans=self.series,
                    scans_projections_indexes=scans_projections_indexes,
                    flip_ud_arr=self.configuration.flip_ud,
                    flip_lr_arr=self.configuration.flip_lr,
                    reading_orders=self.reading_orders,
                ):
                    if self.configuration.rescale_frames:
                        data_frames = self.rescale_frames(data_frames)
                    if self.configuration.normalization_by_sample.is_active():
                        data_frames = self.normalize_frame_by_sample(data_frames)

                    sf = SingleAxisStitcher.stitch_frames(
                        frames=data_frames,
                        axis=self.axis,
                        x_relative_shifts=self._axis_1_rel_final_shifts,
                        y_relative_shifts=self._axis_0_rel_final_shifts,
                        overlap_kernels=self._overlap_kernels,
                        i_frame=i_proj,
                        output_dtype=output_dtype,
                        dumper=self.dumper,
                        return_composition_cls=store_composition if i_proj == 0 else False,
                        stitching_axis=self.axis,
                        pad_mode=self.configuration.pad_mode,
                        alignment=self.configuration.alignment_axis_2,
                        new_width=self._stitching_constant_length,
                        check_inputs=i_proj == 0,  # on process check on the first iteration
                    )
                    if i_proj == 0 and store_composition:
                        _, self._frame_composition = sf
                    if self.progress is not None:
                        self.progress.update()

                    i_proj += 1

            # create link to this dataset that can be missing
            # "data/data" link
            if "data" in h5f[self.configuration.output_data_path]:
                data_group = h5f[self.configuration.output_data_path]["data"]
                if not stitched_frame_path.startswith("/"):
                    stitched_frame_path = "/" + stitched_frame_path
                data_group["data"] = h5py.SoftLink(stitched_frame_path)
                if "default" not in h5f[self.configuration.output_data_path].attrs:
                    h5f[self.configuration.output_data_path].attrs["default"] = "data"
                for attr_name, attr_value in zip(
                    ("NX_class", "SILX_style/axis_scale_types", "signal"),
                    ("NXdata", ["linear", "linear"], "data"),
                ):
                    if attr_name not in data_group.attrs:
                        data_group.attrs[attr_name] = attr_value

        return nx_tomo

    def _create_stitching(self, store_composition):
        self._create_nx_tomo(store_composition=store_composition)

    @staticmethod
    def get_bunch_of_data(
        bunch_start: int,
        bunch_end: int,
        step: int,
        scans: tuple,
        scans_projections_indexes: tuple,
        reading_orders: tuple,
        flip_lr_arr: tuple,
        flip_ud_arr: tuple,
    ):
        """
        goal is to load contiguous projections as much as possible...

        :param int bunch_start: begining of the bunch
        :param int bunch_end: end of the bunch
        :param int scans: ordered scan for which we want to get data
        :param scans_projections_indexes: tuple with scans and scan projection indexes to be loaded
        :param tuple flip_lr_arr: extra information from the user to left-right flip frames
        :param tuple flip_ud_arr: extra information from the user to up-down flip frames
        :return: list of list. For each frame we want to stitch contains the (flat fielded) frames to stich together
        """
        assert len(scans) == len(scans_projections_indexes)
        assert isinstance(flip_lr_arr, tuple)
        assert isinstance(flip_ud_arr, tuple)
        assert isinstance(step, int)
        scans_proj_urls = []
        # for each scan store the real indices and the data url

        for scan, scan_projection_indexes in zip(scans, scans_projections_indexes):
            scan_proj_urls = {}
            # for each scan get the list of url to be loaded
            for i_proj in range(bunch_start, bunch_end):
                if i_proj % step != 0:
                    continue
                proj_index_in_full_scan = scan_projection_indexes[i_proj]
                scan_proj_urls[proj_index_in_full_scan] = scan.projections[proj_index_in_full_scan]
            scans_proj_urls.append(scan_proj_urls)

        # then load data
        all_scan_final_data = numpy.empty((bunch_end - bunch_start, len(scans)), dtype=object)
        from nabu.preproc.flatfield import FlatFieldArrays

        for i_scan, (scan_urls, scan_flip_lr, scan_flip_ud, reading_order) in enumerate(
            zip(scans_proj_urls, flip_lr_arr, flip_ud_arr, reading_orders)
        ):
            i_frame = 0
            _, set_of_compacted_slices = get_compacted_dataslices(scan_urls, return_url_set=True)
            for _, url in set_of_compacted_slices.items():
                scan = scans[i_scan]
                url = DataUrl(
                    file_path=url.file_path(),
                    data_path=url.data_path(),
                    scheme="silx",
                    data_slice=url.data_slice(),
                )
                raw_radios = get_data(url)[::reading_order]
                radio_indices = url.data_slice()
                if isinstance(radio_indices, slice):
                    step = radio_indices.step if radio_indices is not None else 1
                    radio_indices = numpy.arange(
                        start=radio_indices.start,
                        stop=radio_indices.stop,
                        step=step,
                        dtype=numpy.int16,
                    )

                missing = []
                if len(scan.reduced_flats) == 0:
                    missing = "flats"
                if len(scan.reduced_darks) == 0:
                    missing = "darks"

                if len(missing) > 0:
                    _logger.warning(f"missing {'and'.join(missing)}. Unable to do flat field correction")
                    ff_arrays = None
                    data = raw_radios
                else:
                    has_reduced_metadata = (
                        scan.reduced_flats_infos is not None
                        and len(scan.reduced_flats_infos.machine_electric_current) > 0
                        and scan.reduced_darks_infos is not None
                        and len(scan.reduced_darks_infos.machine_electric_current) > 0
                    )
                    if not has_reduced_metadata:
                        _logger.warning("no metadata about current found. Won't normalize according to machine current")

                    ff_arrays = FlatFieldArrays(
                        radios_shape=(len(radio_indices), scan.dim_2, scan.dim_1),
                        flats=scan.reduced_flats,
                        darks=scan.reduced_darks,
                        radios_indices=radio_indices,
                        radios_srcurrent=scan.electric_current[radio_indices] if has_reduced_metadata else None,
                        flats_srcurrent=(
                            scan.reduced_flats_infos.machine_electric_current if has_reduced_metadata else None
                        ),
                    )
                    # note: we need to cast radios to float 32. Darks and flats are cast to anyway
                    data = ff_arrays.normalize_radios(raw_radios.astype(numpy.float32))

                transformations = list(scans[i_scan].get_detector_transformations(tuple()))
                if scan_flip_lr:
                    transformations.append(DetZFlipTransformation(flip=True))
                if scan_flip_ud:
                    transformations.append(DetYFlipTransformation(flip=True))

                transformation_matrix_det_space = build_matrix(transformations)
                if transformation_matrix_det_space is None or numpy.allclose(
                    transformation_matrix_det_space, numpy.identity(3)
                ):
                    flip_ud = False
                    flip_lr = False
                elif numpy.array_equal(transformation_matrix_det_space, PreProcessingStitching._get_UD_flip_matrix()):
                    flip_ud = True
                    flip_lr = False
                elif numpy.allclose(transformation_matrix_det_space, PreProcessingStitching._get_LR_flip_matrix()):
                    flip_ud = False
                    flip_lr = True
                elif numpy.allclose(
                    transformation_matrix_det_space, PreProcessingStitching._get_UD_AND_LR_flip_matrix()
                ):
                    flip_ud = True
                    flip_lr = True
                else:
                    raise ValueError("case not handled... For now only handle up-down flip as left-right flip")

                for frame in data:
                    if flip_ud:
                        frame = numpy.flipud(frame)
                    if flip_lr:
                        frame = numpy.fliplr(frame)
                    all_scan_final_data[i_frame, i_scan] = frame
                    i_frame += 1

        return all_scan_final_data

    def compute_reduced_flats_and_darks(self):
        """
        make sure reduced dark and flats are existing otherwise compute them
        """
        for scan in self.series:
            try:
                reduced_darks, darks_infos = scan.load_reduced_darks(return_info=True)
            except:
                _logger.info("no reduced dark found. Try to compute them.")
            if reduced_darks in (None, {}):
                reduced_darks, darks_infos = scan.compute_reduced_darks(return_info=True)
                try:
                    # if we don't have write in the folder containing the .nx for example
                    scan.save_reduced_darks(reduced_darks, darks_infos=darks_infos)
                except Exception as e:
                    pass
            scan.set_reduced_darks(reduced_darks, darks_infos=darks_infos)

            try:
                reduced_flats, flats_infos = scan.load_reduced_flats(return_info=True)
            except:
                _logger.info("no reduced flats found. Try to compute them.")
            if reduced_flats in (None, {}):
                reduced_flats, flats_infos = scan.compute_reduced_flats(return_info=True)
                try:
                    # if we don't have write in the folder containing the .nx for example
                    scan.save_reduced_flats(reduced_flats, flats_infos=flats_infos)
                except Exception as e:
                    pass
            scan.set_reduced_flats(reduced_flats, flats_infos=flats_infos)

    @staticmethod
    @cache(maxsize=None)
    def _get_UD_flip_matrix():
        return DetYFlipTransformation(flip=True).as_matrix()

    @staticmethod
    @cache(maxsize=None)
    def _get_LR_flip_matrix():
        return DetZFlipTransformation(flip=True).as_matrix()

    @staticmethod
    @cache(maxsize=None)
    def _get_UD_AND_LR_flip_matrix():
        return numpy.matmul(
            PreProcessingStitching._get_UD_flip_matrix(),
            PreProcessingStitching._get_LR_flip_matrix(),
        )

    @staticmethod
    def _get_bunch_of_data(
        bunch_start: int,
        bunch_end: int,
        step: int,
        scans: tuple,
        scans_projections_indexes: tuple,
        reading_orders: tuple,
        flip_lr_arr: tuple,
        flip_ud_arr: tuple,
    ):
        """
        goal is to load contiguous projections as much as possible...

        :param int bunch_start: begining of the bunch
        :param int bunch_end: end of the bunch
        :param int scans: ordered scan for which we want to get data
        :param scans_projections_indexes: tuple with scans and scan projection indexes to be loaded
        :param tuple flip_lr_arr: extra information from the user to left-right flip frames
        :param tuple flip_ud_arr: extra information from the user to up-down flip frames
        :return: list of list. For each frame we want to stitch contains the (flat fielded) frames to stich together
        """
        assert len(scans) == len(scans_projections_indexes)
        assert isinstance(flip_lr_arr, tuple)
        assert isinstance(flip_ud_arr, tuple)
        assert isinstance(step, int)
        scans_proj_urls = []
        # for each scan store the real indices and the data url

        for scan, scan_projection_indexes in zip(scans, scans_projections_indexes):
            scan_proj_urls = {}
            # for each scan get the list of url to be loaded
            for i_proj in range(bunch_start, bunch_end):
                if i_proj % step != 0:
                    continue
                proj_index_in_full_scan = scan_projection_indexes[i_proj]
                scan_proj_urls[proj_index_in_full_scan] = scan.projections[proj_index_in_full_scan]
            scans_proj_urls.append(scan_proj_urls)

        # then load data
        all_scan_final_data = numpy.empty((bunch_end - bunch_start, len(scans)), dtype=object)
        from nabu.preproc.flatfield import FlatFieldArrays

        for i_scan, (scan_urls, scan_flip_lr, scan_flip_ud, reading_order) in enumerate(
            zip(scans_proj_urls, flip_lr_arr, flip_ud_arr, reading_orders)
        ):
            i_frame = 0
            _, set_of_compacted_slices = get_compacted_dataslices(scan_urls, return_url_set=True)
            for _, url in set_of_compacted_slices.items():
                scan = scans[i_scan]
                url = DataUrl(
                    file_path=url.file_path(),
                    data_path=url.data_path(),
                    scheme="silx",
                    data_slice=url.data_slice(),
                )
                raw_radios = get_data(url)[::reading_order]
                radio_indices = url.data_slice()
                if isinstance(radio_indices, slice):
                    step = radio_indices.step if radio_indices is not None else 1
                    radio_indices = numpy.arange(
                        start=radio_indices.start,
                        stop=radio_indices.stop,
                        step=step,
                        dtype=numpy.int16,
                    )

                missing = []
                if len(scan.reduced_flats) == 0:
                    missing = "flats"
                if len(scan.reduced_darks) == 0:
                    missing = "darks"

                if len(missing) > 0:
                    _logger.warning(f"missing {'and'.join(missing)}. Unable to do flat field correction")
                    ff_arrays = None
                    data = raw_radios
                else:
                    has_reduced_metadata = (
                        scan.reduced_flats_infos is not None
                        and len(scan.reduced_flats_infos.machine_electric_current) > 0
                        and scan.reduced_darks_infos is not None
                        and len(scan.reduced_darks_infos.machine_electric_current) > 0
                    )
                    if not has_reduced_metadata:
                        _logger.warning("no metadata about current found. Won't normalize according to machine current")

                    ff_arrays = FlatFieldArrays(
                        radios_shape=(len(radio_indices), scan.dim_2, scan.dim_1),
                        flats=scan.reduced_flats,
                        darks=scan.reduced_darks,
                        radios_indices=radio_indices,
                        radios_srcurrent=scan.electric_current[radio_indices] if has_reduced_metadata else None,
                        flats_srcurrent=(
                            scan.reduced_flats_infos.machine_electric_current if has_reduced_metadata else None
                        ),
                    )
                    # note: we need to cast radios to float 32. Darks and flats are cast to anyway
                    data = ff_arrays.normalize_radios(raw_radios.astype(numpy.float32))

                transformations = list(scans[i_scan].get_detector_transformations(tuple()))
                if scan_flip_lr:
                    transformations.append(DetZFlipTransformation(flip=True))
                if scan_flip_ud:
                    transformations.append(DetYFlipTransformation(flip=True))

                transformation_matrix_det_space = build_matrix(transformations)
                if transformation_matrix_det_space is None or numpy.allclose(
                    transformation_matrix_det_space, numpy.identity(3)
                ):
                    flip_ud = False
                    flip_lr = False
                elif numpy.array_equal(transformation_matrix_det_space, PreProcessingStitching._get_UD_flip_matrix()):
                    flip_ud = True
                    flip_lr = False
                elif numpy.allclose(transformation_matrix_det_space, PreProcessingStitching._get_LR_flip_matrix()):
                    flip_ud = False
                    flip_lr = True
                elif numpy.allclose(
                    transformation_matrix_det_space, PreProcessingStitching._get_UD_AND_LR_flip_matrix()
                ):
                    flip_ud = True
                    flip_lr = True
                else:
                    raise ValueError("case not handled... For now only handle up-down flip as left-right flip")

                for frame in data:
                    if flip_ud:
                        frame = numpy.flipud(frame)
                    if flip_lr:
                        frame = numpy.fliplr(frame)
                    all_scan_final_data[i_frame, i_scan] = frame
                    i_frame += 1

        return all_scan_final_data



from math import ceil
from tempfile import TemporaryDirectory
from dataclasses import dataclass
from tomoscan.io import HDF5File
import pytest
import numpy as np
from nxtomo.application.nxtomo import ImageKey
from tomoscan.esrf import EDFVolume
from nabu.pipeline.reader import NXTomoReaderBinning
from nabu.testutils import utilstest, __do_long_tests__, get_file
from nabu.utils import indices_to_slices, merge_slices
from nabu.io.reader import EDFStackReader, NXTomoReader, NXDarksFlats


@dataclass
class SimpleNXTomoDescription:
    n_darks: int = 0
    n_flats1: int = 0
    n_projs: int = 0
    n_flats2: int = 0
    n_align: int = 0
    frame_shape: tuple = None
    dtype: np.dtype = np.uint16


@pytest.fixture(scope="class")
def bootstrap_nx_reader(request):
    cls = request.cls

    cls.nx_fname = utilstest.getfile("dummy_nxtomo.nx")
    cls.nx_data_path = "entry/instrument/detector/data"
    cls.data_desc = SimpleNXTomoDescription(
        n_darks=10, n_flats1=11, n_projs=100, n_flats2=11, n_align=12, frame_shape=(11, 10), dtype=np.uint16
    )
    cls.projs_vals = np.arange(cls.data_desc.n_projs) + cls.data_desc.n_flats1 + cls.data_desc.n_darks
    cls.darks_vals = np.arange(cls.data_desc.n_darks)
    cls.flats1_vals = np.arange(cls.data_desc.n_darks, cls.data_desc.n_darks + cls.data_desc.n_flats1)
    cls.flats2_vals = np.arange(cls.data_desc.n_darks, cls.data_desc.n_darks + cls.data_desc.n_flats2)

    yield
    # teardown


@pytest.mark.usefixtures("bootstrap_nx_reader")
class TestNXReader:
    def test_incorrect_path(self):
        with pytest.raises(FileNotFoundError):
            reader = NXTomoReader("/invalid/path", self.nx_data_path)
        with pytest.raises(KeyError):
            reader = NXTomoReader(self.nx_fname, "/bad/data/path")

    def test_simple_reads(self):
        """
        Test NXTomoReader with simplest settings
        """
        reader1 = NXTomoReader(self.nx_fname, self.nx_data_path)
        data1 = reader1.load_data()
        assert data1.shape == (self.data_desc.n_projs,) + self.data_desc.frame_shape
        assert np.allclose(data1[:, 0, 0], self.projs_vals)

    def test_image_key(self):
        """
        Test the data selection using "image_key".
        """
        reader_projs = NXTomoReader(self.nx_fname, self.nx_data_path, image_key=ImageKey.PROJECTION.value)
        data = reader_projs.load_data()
        assert np.allclose(data[:, 0, 0], self.projs_vals)

        reader_darks = NXTomoReader(self.nx_fname, self.nx_data_path, image_key=ImageKey.DARK_FIELD.value)
        data_darks = reader_darks.load_data()
        assert np.allclose(data_darks[:, 0, 0], self.darks_vals)

        reader_flats = NXTomoReader(self.nx_fname, self.nx_data_path, image_key=ImageKey.FLAT_FIELD.value)
        data_flats = reader_flats.load_data()
        assert np.allclose(data_flats[:, 0, 0], np.concatenate([self.flats1_vals, self.flats2_vals]))

    def test_data_buffer_and_subregion(self):
        """
        Test the "data_buffer" and "sub_region" parameters
        """
        data_desc = self.data_desc

        def _check_correct_shape_succeeds(shape, sub_region, test_description=""):
            err_msg = "Something wrong with the following test:" + test_description
            data_buffer = np.zeros(shape, dtype="f")
            reader1 = NXTomoReader(self.nx_fname, self.nx_data_path, sub_region=sub_region)
            data1 = reader1.load_data(output=data_buffer)
            assert id(data1) == id(data_buffer), err_msg
            reader2 = NXTomoReader(self.nx_fname, self.nx_data_path, sub_region=sub_region)
            data2 = reader2.load_data()
            assert np.allclose(data1, data2), err_msg

        test_cases = [
            {
                "description": "In the projections, read everything into the provided data buffer",
                "sub_region": None,
                "correct_shape": (data_desc.n_projs,) + data_desc.frame_shape,
                "wrong_shapes": [
                    (data_desc.n_projs - 1,) + data_desc.frame_shape,
                    (data_desc.n_projs - 1,) + (999, 998),
                    (data_desc.n_projs,) + (999, 998),
                ],
            },
            {
                "description": "In the projections, select a subset along dimension 0 (i.e take only several full frames). The correct output shape is: data_total[image_key==0][slice(10, 30)].shape",
                "sub_region": slice(10, 30),
                "correct_shape": (20,) + data_desc.frame_shape,
                "wrong_shapes": [
                    (data_desc.n_projs,) + data_desc.frame_shape,
                    (19,) + data_desc.frame_shape,
                ],
            },
            {
                "description": "In the projections, read several rows of all images, i.e extract several sinograms. The correct output shape is: data_total[image_key==0][:, slice(start_z, end_z), :].shape",
                "sub_region": (None, slice(3, 7), None),
                "correct_shape": (data_desc.n_projs, 4, data_desc.frame_shape[-1]),
                "wrong_shapes": [],
            },
        ]

        for test_case in test_cases:
            for wrong_shape in test_case["wrong_shapes"]:
                with pytest.raises(ValueError):
                    data_buffer_wrong_shape = np.zeros(wrong_shape, dtype="f")
                    reader = NXTomoReader(
                        self.nx_fname,
                        self.nx_data_path,
                        sub_region=test_case["sub_region"],
                    )
                    reader.load_data(output=data_buffer_wrong_shape)
            _check_correct_shape_succeeds(test_case["correct_shape"], test_case["sub_region"], test_case["description"])

    def test_subregion_and_subsampling(self):
        data_desc = self.data_desc
        test_cases = [
            {
                # Read one full image out of two in all projections
                "sub_region": (slice(None, None, 2), None, None),
                "expected_shape": (self.projs_vals[::2].size,) + data_desc.frame_shape,
                "expected_values": self.projs_vals[::2],
            },
            {
                # Read one image fragment (several rows) out of two in all projections
                "sub_region": (slice(None, None, 2), slice(5, 8), None),
                "expected_shape": (self.projs_vals[::2].size, 3, data_desc.frame_shape[-1]),
                "expected_values": self.projs_vals[::2],
            },
        ]

        for test_case in test_cases:
            reader = NXTomoReader(self.nx_fname, self.nx_data_path, sub_region=test_case["sub_region"])
            data = reader.load_data()
            assert data.shape == test_case["expected_shape"]
            assert np.allclose(data[:, 0, 0], test_case["expected_values"])

    def test_reading_with_binning_(self):
        from nabu.pipeline.reader import NXTomoReaderBinning

        reader_with_binning = NXTomoReaderBinning((2, 2), self.nx_fname, self.nx_data_path)
        data = reader_with_binning.load_data()
        assert data.shape == (self.data_desc.n_projs,) + tuple(n // 2 for n in self.data_desc.frame_shape)

    def test_reading_with_distortion_correction(self):
        from nabu.io.detector_distortion import DetectorDistortionBase
        from nabu.pipeline.reader import NXTomoReaderDistortionCorrection

        data_desc = self.data_desc

        # (start_x, end_x, start_y, end_y)
        sub_region_xy = (None, None, 1, 6)

        distortion_corrector = DetectorDistortionBase(detector_full_shape_vh=data_desc.frame_shape)
        distortion_corrector.set_sub_region_transformation(target_sub_region=sub_region_xy)
        adapted_subregion = distortion_corrector.get_adapted_subregion(sub_region_xy)
        sub_region = (slice(None, None), slice(*sub_region_xy[2:]), slice(*sub_region_xy[:2]))

        reader_distortion_corr = NXTomoReaderDistortionCorrection(
            distortion_corrector,
            self.nx_fname,
            self.nx_data_path,
            sub_region=sub_region,
        )

        reader_distortion_corr.load_data()

    @pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1")
    def test_other_load_patterns(self):
        """
        Other data read patterns that are sometimes used by ChunkedPipeline
        Test cases already done in check_correct_shape_succeeds():
            - Read all frames in a provided buffer
            - Read a subset of all (full) projections
            - Read several rows of all projections (extract sinograms)
        """
        data_desc = self.data_desc

        test_cases = [
            {
                "description": "Select a subset along all dimensions. The correct output shape is data_total[image_key==0][slice_dim0, slice_dim1, slice_dim2].shape",
                "sub_region": (slice(10, 72, 2), slice(4, None), slice(2, 8)),
                "expected_shape": (31, 7, 6),
                "expected_values": self.projs_vals[slice(10, 72, 2)],
            },
            {
                "description": "Select several rows in all images (i.e extract sinograms), with binning",
                "sub_region": (slice(None, None), slice(3, 7), slice(None, None)),
                "binning": (2, 2),
                "expected_shape": (data_desc.n_projs, 4 // 2, data_desc.frame_shape[-1] // 2),
                "expected_values": self.projs_vals[:],
            },
            {
                "description": "Extract sinograms with binning + subsampling",
                "sub_region": (slice(None, None, 2), slice(1, 8), slice(None, None)),
                "binning": (2, 2),
                "expected_shape": (ceil(data_desc.n_projs / 2), 7 // 2, data_desc.frame_shape[-1] // 2),
                "expected_values": self.projs_vals[::2],
            },
        ]

        for test_case in test_cases:
            binning = test_case.get("binning", None)
            reader_cls = NXTomoReader
            init_args = [self.nx_fname, self.nx_data_path]
            init_kwargs = {"sub_region": test_case["sub_region"]}
            if binning is not None:
                reader_cls = NXTomoReaderBinning
                init_args = [binning] + init_args
            reader = reader_cls(*init_args, **init_kwargs)
            data = reader.load_data()
            err_msg = "Something wrong with test: " + test_case["description"]
            assert data.shape == test_case["expected_shape"], err_msg
            assert np.allclose(data[:, 0, 0], test_case["expected_values"]), err_msg


@pytest.fixture(scope="class")
def bootstrap_edf_reader(request):
    cls = request.cls

    test_dir = utilstest.data_home
    cls._tmpdir = TemporaryDirectory(prefix="test_edf_stack_", dir=test_dir)
    cls.edf_dir = cls._tmpdir.name
    cls.n_projs = 100
    cls.frame_shape = (11, 12)
    cls.projs_vals = np.arange(cls.n_projs, dtype=np.uint16) + 10

    edf_vol = EDFVolume(folder=cls.edf_dir, volume_basename="edf_stack", overwrite=True)
    data_shape = (cls.n_projs,) + cls.frame_shape
    edf_vol.data = np.ones(data_shape, dtype=np.uint16) * cls.projs_vals.reshape(cls.n_projs, 1, 1)
    edf_vol.save_data()
    cls.filenames = list(edf_vol.browse_data_files())

    yield
    cls._tmpdir.cleanup()


@pytest.mark.usefixtures("bootstrap_edf_reader")
class TestEDFReader:
    def test_read_all_frames(self):
        """
        Simple test, read all the frames
        """
        reader = EDFStackReader(self.filenames)
        data = reader.load_data()
        expected_shape = (self.n_projs,) + self.frame_shape
        assert data.shape == expected_shape
        assert np.allclose(data[:, 0, 0], self.projs_vals)

        buffer_correct = np.zeros(expected_shape, dtype=np.float32)
        reader.load_data(output=buffer_correct)

        buffer_incorrect_1 = np.zeros((99, 11, 12), dtype=np.float32)
        with pytest.raises(ValueError):
            reader.load_data(output=buffer_incorrect_1)

        buffer_incorrect_2 = np.zeros((100, 11, 12), dtype=np.uint16)
        with pytest.raises(ValueError):
            reader.load_data(output=buffer_incorrect_2)

    def test_subregions_1(self):
        test_cases = [
            {
                "name": "read a handful of full frames",
                "sub_region": (slice(0, 48), slice(None, None), slice(None, None)),
                "expected_shape": (48,) + self.frame_shape,
                "expected_values": self.projs_vals[:48],
            },
            {
                "name": "read several lines of all frames (i.e extract a singoram)",
                "sub_region": (slice(None, None), slice(0, 6), slice(None, None)),
                "expected_shape": (self.n_projs, 6, self.frame_shape[-1]),
                "expected_values": self.projs_vals,
            },
            {
                "name": "read several lines of all frames (i.e extract a singoram), and a X-ROI",
                "sub_region": (slice(None, None), slice(3, 7), slice(2, 5)),
                "expected_shape": (self.n_projs, 4, 3),
                "expected_values": self.projs_vals,
            },
            {
                "name": "read several lines of all frames (i.e extract a singoram), with angular subsampling",
                "sub_region": (slice(None, None, 2), slice(3, 7), slice(2, 5)),
                "expected_shape": (ceil(self.n_projs / 2), 4, 3),
                "expected_values": self.projs_vals[::2],
            },
        ]
        for test_case in test_cases:
            reader = EDFStackReader(self.filenames, sub_region=test_case["sub_region"])
            data = reader.load_data()
            err_msg = "Something wrong with test: %s" % (test_case["name"])
            assert data.shape == test_case["expected_shape"], err_msg
            assert np.allclose(data[:, 0, 0], test_case["expected_values"]), err_msg

    @pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1")
    def test_reading_with_binning(self):
        from nabu.pipeline.reader import EDFStackReaderBinning

        reader_with_binning = EDFStackReaderBinning((2, 2), self.filenames)
        data = reader_with_binning.load_data()
        assert data.shape == (self.n_projs,) + tuple(n // 2 for n in self.frame_shape)

    @pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1")
    def test_reading_with_distortion_correction(self):
        from nabu.io.detector_distortion import DetectorDistortionBase
        from nabu.pipeline.reader import EDFStackReaderDistortionCorrection

        # (start_x, end_x, start_y, end_y)
        sub_region_xy = (None, None, 1, 6)

        distortion_corrector = DetectorDistortionBase(detector_full_shape_vh=self.frame_shape)
        distortion_corrector.set_sub_region_transformation(target_sub_region=sub_region_xy)
        adapted_subregion = distortion_corrector.get_adapted_subregion(sub_region_xy)
        sub_region = (slice(None, None), slice(*sub_region_xy[2:]), slice(*sub_region_xy[:2]))

        reader_distortion_corr = EDFStackReaderDistortionCorrection(
            distortion_corrector,
            self.filenames,
            sub_region=sub_region,
        )

        reader_distortion_corr.load_data()


def test_indices_to_slices():
    slices1 = [slice(0, 4)]
    slices2 = [slice(11, 16)]
    slices3 = [slice(3, 5), slice(8, 20)]
    slices4 = [slice(2, 7), slice(18, 28), slice(182, 845)]
    idx = np.arange(1000)
    for slices in [slices1, slices2, slices3, slices4]:
        indices = np.hstack([idx[sl] for sl in slices])
        slices_calculated = indices_to_slices(indices)
        assert slices_calculated == slices, "Expected indices_to_slices() to return %s, but got %s" % (
            str(slices),
            str(slices_calculated),
        )


def test_merge_slices():
    idx = np.arange(10000)
    rnd = lambda x: np.random.randint(1, high=x)

    n_tests = 10
    for i in range(n_tests):
        start1 = rnd(1000)
        stop1 = start1 + rnd(1000)
        start2 = rnd(1000)
        stop2 = start2 + rnd(1000)
        step1 = rnd(4)
        step2 = rnd(4)
        slice1 = slice(start1, stop1, step1)
        slice2 = slice(start2, stop2, step2)

        assert np.allclose(idx[slice1][slice2], idx[merge_slices(slice1, slice2)])


@pytest.fixture(scope="class")
def bootstrap_nxdkrf(request):
    cls = request.cls

    cls.nx_file_path = get_file("bamboo_reduced.nx")

    yield
    # teardown


@pytest.mark.usefixtures("bootstrap_nxdkrf")
class TestDKRFReader:
    def test_darks(self):
        dkrf_reader = NXDarksFlats(self.nx_file_path)
        darks = dkrf_reader.get_raw_darks(as_multiple_array=True)
        reduced_darks = dkrf_reader.get_reduced_darks(method="mean")

        actual_darks = []
        with HDF5File(self.nx_file_path, "r") as f:
            actual_darks.append(f["entry0000/data/data"][slice(0, 1)])

        assert len(darks) == len(actual_darks)

        for i in range(len(darks)):
            assert np.allclose(darks[i], actual_darks[i])
            actual_reduced_darks = np.mean(actual_darks[i], axis=0)
            assert np.allclose(reduced_darks[i], actual_reduced_darks)

        assert np.allclose(list(dkrf_reader.get_reduced_darks(as_dict=True).keys()), [0])

    def test_flats(self):
        dkrf_reader = NXDarksFlats(self.nx_file_path)
        flats = dkrf_reader.get_raw_flats(as_multiple_array=True)
        reduced_flats = dkrf_reader.get_reduced_flats(method="median")

        actual_flats = []
        with HDF5File(self.nx_file_path, "r") as f:
            actual_flats.append(f["entry0000/data/data"][slice(1, 25 + 1)])
            actual_flats.append(f["entry0000/data/data"][slice(526, 550 + 1)])

        assert len(flats) == len(actual_flats)
        for i in range(len(flats)):
            assert np.allclose(flats[i], actual_flats[i])
            actual_reduced_flats = np.median(actual_flats[i], axis=0)
            assert np.allclose(reduced_flats[i], actual_reduced_flats)

        assert np.allclose(list(dkrf_reader.get_reduced_flats(as_dict=True).keys()), [1, 526])



import h5py
import numpy
import logging
from .base import DumperBase
from nabu.stitching.config import PreProcessedSingleAxisStitchingConfiguration
from nabu import version as nabu_version
from nabu.io.writer import get_datetime
from silx.io.dictdump import dicttonx
from tomoscan.identifier import ScanIdentifier


_logger = logging.getLogger(__name__)


class PreProcessingStitchingDumper(DumperBase):
    """
    dumper to be used when save data durint pre-processing stitching (on projections). Output is expected to be an NXtomo
    """

    def __init__(self, configuration) -> None:
        if not isinstance(configuration, PreProcessedSingleAxisStitchingConfiguration):
            raise TypeError(
                f"configuration is expected to be an instance of {PreProcessedSingleAxisStitchingConfiguration}. Get {type(configuration)} instead"
            )
        super().__init__(configuration)

    def save_frame_to_disk(self, output_dataset: h5py.Dataset, index: int, stitched_frame: numpy.ndarray, **kwargs):
        output_dataset[index] = stitched_frame

    def save_configuration(self):
        """dump configuration used for stitching at the NXtomo entry"""
        process_name = "stitching_configuration"
        config_dict = self.configuration.to_dict()
        # adding nabu specific information
        nabu_process_info = {
            "@NX_class": "NXentry",
            f"{process_name}@NX_class": "NXprocess",
            f"{process_name}/program": "nabu-stitching",
            f"{process_name}/version": nabu_version,
            f"{process_name}/date": get_datetime(),
            f"{process_name}/configuration": config_dict,
        }

        dicttonx(
            nabu_process_info,
            h5file=self.configuration.output_file_path,
            h5path=self.configuration.output_data_path,
            update_mode="replace",
            mode="a",
        )

    @property
    def output_identifier(self) -> ScanIdentifier:
        return self.configuration.get_output_object().get_identifier()

    def create_output_dataset(self):
        """
        function called at the beginning of the stitching to prepare output dataset
        """
        raise NotImplementedError



import h5py
import numpy
import logging
from math import ceil
from typing import Optional, Iterable, Union
from tomoscan.series import Series
from tomoscan.identifier import BaseIdentifier
from nabu.stitching.stitcher.base import _StitcherBase, get_obj_constant_side_length
from nabu.stitching.stitcher_2D import stitch_raw_frames
from nabu.stitching.utils.utils import ShiftAlgorithm, from_slice_to_n_elements
from nabu.stitching.overlap import (
    check_overlaps,
    ImageStichOverlapKernel,
)
from nabu.stitching.config import (
    SingleAxisStitchingConfiguration,
    KEY_RESCALE_MIN_PERCENTILES,
    KEY_RESCALE_MAX_PERCENTILES,
)
from nabu.misc.utils import rescale_data
from nabu.stitching.sample_normalization import normalize_frame as normalize_frame_by_sample
from nabu.stitching.stitcher.dumper.base import DumperBase
from silx.io.utils import get_data
from silx.io.url import DataUrl
from scipy.ndimage import shift as shift_scipy


_logger = logging.getLogger(__name__)


PROGRESS_BAR_STITCH_VOL_DESC = "stitch volumes"
# description of the progress bar used when stitching volume.
# Needed to retrieve advancement from file when stitching remotely


class _SingleAxisMetaClass(type):
    """
    Metaclass for single axis stitcher in order to aggregate dumper class and axis
    """

    def __new__(mcls, name, bases, attrs, axis=None, dumper_cls=None):
        mcls = super().__new__(mcls, name, bases, attrs)
        mcls._axis = axis
        mcls._dumperCls = dumper_cls
        return mcls


class SingleAxisStitcher(_StitcherBase, metaclass=_SingleAxisMetaClass):
    """
    Any single-axis base class
    """

    def __init__(self, configuration, *args, **kwargs) -> None:
        super().__init__(configuration, *args, **kwargs)
        if self._dumperCls is not None:
            self._dumper = self._dumperCls(configuration=configuration)
        else:
            self._dumper = None

        # initial shifts
        self._axis_0_rel_ini_shifts = []
        """Shift between two juxtapose objects along axis 0 found from position metadata or given by the user"""
        self._axis_1_rel_ini_shifts = []
        """Shift between two juxtapose objects along axis 1 found from position metadata or given by the user"""
        self._axis_2_rel_ini_shifts = []
        """Shift between two juxtapose objects along axis 2 found from position metadata or given by the user"""

        # shifts to add once refine
        self._axis_0_rel_final_shifts = []
        """Shift over axis 0 found once refined by the cross correlation algorithm"""
        self._axis_1_rel_final_shifts = []
        """Shift over axis 1 found once refined by the cross correlation algorithm"""
        self._axis_2_rel_final_shifts = []
        """Shift over axis 2 found once refined by the cross correlation algorithm"""

        self._slices_to_stitch = None
        # slices to be stitched. Obtained from calling Configuration.settle_slices

        self._stitching_constant_length = None
        # stitching width: larger volume width. Other volume will be pad

        def shifts_is_scalar(shifts):
            return isinstance(shifts, ShiftAlgorithm) or numpy.isscalar(shifts)

        # 'expend' shift algorithm
        if shifts_is_scalar(self.configuration.axis_0_pos_px):
            self.configuration.axis_0_pos_px = [
                self.configuration.axis_0_pos_px,
            ] * (len(self.series) - 1)
        if shifts_is_scalar(self.configuration.axis_1_pos_px):
            self.configuration.axis_1_pos_px = [
                self.configuration.axis_1_pos_px,
            ] * (len(self.series) - 1)
        if shifts_is_scalar(self.configuration.axis_1_pos_px):
            self.configuration.axis_1_pos_px = [
                self.configuration.axis_1_pos_px,
            ] * (len(self.series) - 1)
        if numpy.isscalar(self.configuration.axis_0_params):
            self.configuration.axis_0_params = [
                self.configuration.axis_0_params,
            ] * (len(self.series) - 1)
        if numpy.isscalar(self.configuration.axis_1_params):
            self.configuration.axis_1_params = [
                self.configuration.axis_1_params,
            ] * (len(self.series) - 1)
        if numpy.isscalar(self.configuration.axis_2_params):
            self.configuration.axis_2_params = [
                self.configuration.axis_2_params,
            ] * (len(self.series) - 1)

    @property
    def axis(self) -> int:
        return self._axis

    @property
    def dumper(self):
        return self._dumper

    @property
    def stitching_axis_in_frame_space(self):
        """
        stitching is operated in 2D (frame) space. So the axis in frame space is different than the one in 3D ebs-tomo space (https://tomo.gitlab-pages.esrf.fr/bliss-tomo/master/modelization.html)
        """
        raise NotImplementedError("Base class")

    def stitch(self, store_composition: bool = True) -> BaseIdentifier:
        if self.progress is not None:
            self.progress.set_description("order scans")
        self.order_input_tomo_objects()
        if self.progress is not None:
            self.progress.set_description("check inputs")
        self.check_inputs()
        self.settle_flips()

        if self.progress is not None:
            self.progress.set_description("compute shifts")
        self._compute_positions_as_px()
        self.pre_processing_computation()

        self.compute_estimated_shifts()
        self._compute_shifts()
        self._createOverlapKernels()
        if self.progress is not None:
            self.progress.set_description(PROGRESS_BAR_STITCH_VOL_DESC)

        self._create_stitching(store_composition=store_composition)
        if self.progress is not None:
            self.progress.set_description("dump configuration")
        self.dumper.save_configuration()
        return self.dumper.output_identifier

    @property
    def serie_label(self) -> str:
        """return serie name for logs"""
        return "single axis serie"

    def get_n_slices_to_stitch(self):
        """Return the number of slice to be stitched"""
        if self._slices_to_stitch is None:
            raise RuntimeError("Slices needs to be settled first")
        return from_slice_to_n_elements(self._slices_to_stitch)

    def get_final_axis_positions_in_px(self) -> dict:
        """
        compute the final position (**in pixel**) from the initial position of the first object and the final relative shift computed (1)
        (1): the final relative shift is obtained from the initial shift (from motor position of provided by the user) + the refinement shift from cross correlation algorithm
        :return: dict with tomo object identifier (str) as key and a tuple of position in pixel (axis_0_pos, axis_1_pos, axis_2_pos)
        """
        pos_0_shift = numpy.concatenate(
            (
                numpy.atleast_1d(0.0),
                numpy.array(self._axis_0_rel_final_shifts) - numpy.array(self._axis_0_rel_ini_shifts),
            )
        )
        pos_0_cum_shift = numpy.cumsum(pos_0_shift)
        final_pos_axis_0 = self.configuration.axis_0_pos_px + pos_0_cum_shift

        pos_1_shift = numpy.concatenate(
            (
                numpy.atleast_1d(0.0),
                numpy.array(self._axis_1_rel_final_shifts) - numpy.array(self._axis_1_rel_ini_shifts),
            )
        )
        pos_1_cum_shift = numpy.cumsum(pos_1_shift)
        final_pos_axis_1 = self.configuration.axis_1_pos_px + pos_1_cum_shift

        pos_2_shift = numpy.concatenate(
            (
                numpy.atleast_1d(0.0),
                numpy.array(self._axis_2_rel_final_shifts) - numpy.array(self._axis_2_rel_ini_shifts),
            )
        )
        pos_2_cum_shift = numpy.cumsum(pos_2_shift)
        final_pos_axis_2 = self.configuration.axis_2_pos_px + pos_2_cum_shift

        assert len(final_pos_axis_0) == len(final_pos_axis_1)
        assert len(final_pos_axis_0) == len(final_pos_axis_2)
        assert len(final_pos_axis_0) == len(self.series)

        return {
            tomo_obj.get_identifier().to_str(): (pos_0, pos_1, pos_2)
            for tomo_obj, (pos_0, pos_1, pos_2) in zip(
                self.series, zip(final_pos_axis_0, final_pos_axis_1, final_pos_axis_2)
            )
        }

    def settle_flips(self):
        """
        User can provide some information on existing flips at frame level.
        The goal of this step is to get one flip_lr and on flip_ud value per scan or volume
        """
        if numpy.isscalar(self.configuration.flip_lr):
            self.configuration.flip_lr = tuple([self.configuration.flip_lr] * len(self.series))
        else:
            if not len(self.configuration.flip_lr) == len(self.series):
                raise ValueError("flip_lr expects a scalar value or one value per element to stitch")
            self.configuration.flip_lr = tuple(self.configuration.flip_lr)
            for elmt in self.configuration.flip_lr:
                if not isinstance(elmt, bool):
                    raise TypeError

        if numpy.isscalar(self.configuration.flip_ud):
            self.configuration.flip_ud = tuple([self.configuration.flip_ud] * len(self.series))
        else:
            if not len(self.configuration.flip_ud) == len(self.series):
                raise ValueError("flip_ud expects a scalar value or one value per element to stitch")
            self.configuration.flip_ud = tuple(self.configuration.flip_ud)
            for elmt in self.configuration.flip_ud:
                if not isinstance(elmt, bool):
                    raise TypeError

    def _createOverlapKernels(self):
        """
        after this stage the overlap kernels must be created and with the final overlap size
        """
        if self.axis == 0:
            stitched_axis_rel_shifts = self._axis_0_rel_final_shifts
            stitched_axis_params = self.configuration.axis_0_params
        elif self.axis == 1:
            stitched_axis_rel_shifts = self._axis_1_rel_final_shifts
            stitched_axis_params = self.configuration.axis_1_params
        elif self.axis == 2:
            stitched_axis_rel_shifts = self._axis_2_rel_final_shifts
            stitched_axis_params = self.configuration.axis_2_params
        else:
            raise NotImplementedError

        if stitched_axis_rel_shifts is None or len(stitched_axis_rel_shifts) == 0:
            raise RuntimeError(
                f"axis {self.axis} shifts have not been defined yet. Please define them before calling this function"
            )

        overlap_size = stitched_axis_params.get("overlap_size", None)
        if overlap_size in (None, "None", ""):
            overlap_size = -1
        else:
            overlap_size = int(overlap_size)

        self._stitching_constant_length = max(
            [get_obj_constant_side_length(obj, axis=self.axis) for obj in self.series]
        )

        for stitched_axis_shift in stitched_axis_rel_shifts:
            if overlap_size == -1:
                height = abs(stitched_axis_shift)
            else:
                height = overlap_size

            self._overlap_kernels.append(
                ImageStichOverlapKernel(
                    stitching_axis=self.stitching_axis_in_frame_space,
                    frame_unstitched_axis_size=self._stitching_constant_length,
                    stitching_strategy=self.configuration.stitching_strategy,
                    overlap_size=height,
                    extra_params=self.configuration.stitching_kernels_extra_params,
                )
            )

    @property
    def series(self) -> Series:
        return self._series

    @property
    def configuration(self) -> SingleAxisStitchingConfiguration:
        return self._configuration

    @property
    def progress(self):
        return self._progress

    @staticmethod
    def _data_bunch_iterator(slices, bunch_size):
        """util to get indices by bunch until we reach n_frames"""
        if isinstance(slices, slice):
            # note: slice step is handled at a different level
            start = end = slices.start

            while True:
                start, end = end, min((end + bunch_size), slices.stop)
                yield (start, end)
                if end >= slices.stop:
                    break
        # in the case of non-contiguous frames
        elif isinstance(slices, Iterable):
            for s in slices:
                yield (s, s + 1)
        else:
            raise TypeError(f"slices is provided as {type(slices)}. When Iterable or slice is expected")

    def rescale_frames(self, frames: tuple):
        """
        rescale_frames if requested by the configuration
        """
        _logger.info("apply rescale frames")

        def cast_percentile(percentile) -> int:
            if isinstance(percentile, str):
                percentile.replace(" ", "").rstrip("%")
            return int(percentile)

        rescale_min_percentile = cast_percentile(self.configuration.rescale_params.get(KEY_RESCALE_MIN_PERCENTILES, 0))
        rescale_max_percentile = cast_percentile(
            self.configuration.rescale_params.get(KEY_RESCALE_MAX_PERCENTILES, 100)
        )

        new_min = numpy.percentile(frames[0], rescale_min_percentile)
        new_max = numpy.percentile(frames[0], rescale_max_percentile)

        def rescale(data):
            # FIXME: takes time because browse several time the dataset, twice for percentiles and twices to get min and max when calling rescale_data...
            data_min = numpy.percentile(data, rescale_min_percentile)
            data_max = numpy.percentile(data, rescale_max_percentile)
            return rescale_data(data, new_min=new_min, new_max=new_max, data_min=data_min, data_max=data_max)

        return tuple([rescale(data) for data in frames])

    def normalize_frame_by_sample(self, frames: tuple):
        """
        normalize frame from a sample picked on the left or the right
        """
        _logger.info("apply normalization by a sample")
        return tuple(
            [
                normalize_frame_by_sample(
                    frame=frame,
                    side=self.configuration.normalization_by_sample.side,
                    method=self.configuration.normalization_by_sample.method,
                    margin_before_sample=self.configuration.normalization_by_sample.margin,
                    sample_width=self.configuration.normalization_by_sample.width,
                )
                for frame in frames
            ]
        )

    @staticmethod
    def stitch_frames(
        frames: Union[tuple, numpy.ndarray],
        axis,
        x_relative_shifts: tuple,
        y_relative_shifts: tuple,
        output_dtype: numpy.ndarray,
        stitching_axis: int,
        overlap_kernels: tuple,
        dumper: DumperBase = None,
        check_inputs=True,
        shift_mode="nearest",
        i_frame=None,
        return_composition_cls=False,
        alignment="center",
        pad_mode="constant",
        new_width: Optional[int] = None,
    ) -> numpy.ndarray:
        """
        shift frames according to provided `shifts` (as y, x tuples) then stitch all the shifted frames together and
        save them to output_dataset.

        :param tuple frames: element must be a DataUrl or a 2D numpy array
        :param stitching_regions_hdf5_dataset:
        """
        if check_inputs:
            if len(frames) < 2:
                raise ValueError(f"Not enought frames provided for stitching ({len(frames)} provided)")
            if len(frames) != len(x_relative_shifts) + 1:
                raise ValueError(
                    f"Incoherent number of shift provided ({len(x_relative_shifts)}) compare to number of frame ({len(frames)}). len(frames) - 1 expected"
                )
            if len(x_relative_shifts) != len(overlap_kernels):
                raise ValueError(
                    f"expect to have the same number of x_relative_shifts ({len(x_relative_shifts)}) and y_overlap ({len(overlap_kernels)})"
                )
            if len(y_relative_shifts) != len(overlap_kernels):
                raise ValueError(
                    f"expect to have the same number of y_relative_shifts ({len(y_relative_shifts)}) and y_overlap ({len(overlap_kernels)})"
                )

            relative_positions = [(0, 0, 0)]
            for y_rel_pos, x_rel_pos in zip(y_relative_shifts, x_relative_shifts):
                relative_positions.append(
                    (
                        y_rel_pos + relative_positions[-1][0],
                        0,  # position over axis 1 (aka y) is not handled yet
                        x_rel_pos + relative_positions[-1][2],
                    )
                )
            check_overlaps(
                frames=tuple(frames),
                positions=tuple(relative_positions),
                axis=axis,
                raise_error=False,
            )

        def check_frame_is_2d(frame):
            if frame.ndim != 2:
                raise ValueError(f"2D frame expected when {frame.ndim}D provided")

        # step_0 load data if from url
        data = []
        for frame in frames:
            if isinstance(frame, DataUrl):
                data_frame = get_data(frame)
                if check_inputs:
                    check_frame_is_2d(data_frame)
                data.append(data_frame)
            elif isinstance(frame, numpy.ndarray):
                if check_inputs:
                    check_frame_is_2d(frame)
                data.append(frame)
            else:
                raise TypeError(f"frames are expected to be DataUrl or 2D numpy array. Not {type(frame)}")

        # step 1: shift each frames (except the first one)
        if stitching_axis == 0:
            relative_shift_along_stitched_axis = y_relative_shifts
            relative_shift_along_unstitched_axis = x_relative_shifts
        elif stitching_axis == 1:
            relative_shift_along_stitched_axis = x_relative_shifts
            relative_shift_along_unstitched_axis = y_relative_shifts
        else:
            raise NotImplementedError("")

        shifted_data = [data[0]]
        for frame, relative_shift in zip(data[1:], relative_shift_along_unstitched_axis):
            # note: for now we only shift data in x. the y shift is handled in the FrameComposition
            relative_shift = numpy.asarray(relative_shift).astype(numpy.int8)
            if relative_shift == 0:
                shifted_frame = frame
            else:
                # TO speed up: should use the Fourier transform
                shifted_frame = shift_scipy(
                    frame,
                    mode=shift_mode,
                    shift=[0, -relative_shift] if stitching_axis == 0 else [-relative_shift, 0],
                    order=1,
                )
            shifted_data.append(shifted_frame)

        # step 2: create stitched frame
        stitched_frame, composition_cls = stitch_raw_frames(
            frames=shifted_data,
            key_lines=(
                [
                    (int(frame.shape[stitching_axis] - abs(relative_shift / 2)), int(abs(relative_shift / 2)))
                    for relative_shift, frame in zip(relative_shift_along_stitched_axis, frames)
                ]
            ),
            overlap_kernels=overlap_kernels,
            check_inputs=check_inputs,
            output_dtype=output_dtype,
            return_composition_cls=True,
            alignment=alignment,
            pad_mode=pad_mode,
            new_unstitched_axis_size=new_width,
        )
        dumper.save_stitched_frame(
            stitched_frame=stitched_frame,
            composition_cls=composition_cls,
            i_frame=i_frame,
            axis=1,
        )

        if return_composition_cls:
            return stitched_frame, composition_cls
        else:
            return stitched_frame



import os
import numpy as np
from silx.io.url import DataUrl
from silx.io import get_data
from tomoscan.esrf.scan.edfscan import EDFTomoScan
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan

from ..utils import check_supported, indices_to_slices
from ..io.reader import EDFStackReader, NXDarksFlats, NXTomoReader
from ..io.utils import get_compacted_dataslices
from .utils import get_values_from_file, is_hdf5_extension
from .logger import LoggerOrPrint

from ..pipeline.utils import nabu_env_settings


class DatasetAnalyzer:
    _scanner = None
    kind = "none"

    """
    Base class for datasets analyzers.
    """

    def __init__(self, location, extra_options=None, logger=None):
        """
        Initialize a Dataset analyzer.

        Parameters
        ----------
        location: str
            Dataset location (directory or file name)
        extra_options: dict, optional
            Extra options on how to interpret the dataset.
        logger: logging object, optional
            Logger. If not set, messages will just be printed in stdout.
        """
        self.logger = LoggerOrPrint(logger)
        self.location = location
        self._set_extra_options(extra_options)
        self._get_excluded_projections()
        self._set_default_dataset_values()
        self._init_dataset_scan()
        self._finish_init()

    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        # COMPAT.
        advanced_options = {
            "force_flatfield": False,
            "output_dir": None,
            "exclude_projections": None,
            "hdf5_entry": None,
            # "nx_version": 1.0,
        }
        # --
        advanced_options.update(extra_options)
        self.extra_options = advanced_options

    # pylint: disable=E1136
    def _get_excluded_projections(self):
        excluded_projs = self.extra_options["exclude_projections"]
        self._ignore_projections = None
        if excluded_projs is None:
            return

        if excluded_projs["type"] == "angular_range":
            excluded_projs["type"] = "range"  # compat with tomoscan #pylint: disable=E1137
            values = excluded_projs["range"]
        for ignore_kind, dtype in {"indices": np.int32, "angles": np.float32}.items():
            if excluded_projs["type"] == ignore_kind:
                values = get_values_from_file(excluded_projs["file"], any_size=True).astype(dtype).tolist()
        self._ignore_projections = {"kind": excluded_projs["type"], "values": values}  # pylint: disable=E0606

    def _init_dataset_scan(self, **kwargs):
        if self._scanner is None:
            raise ValueError("Base class")
        if self._scanner is NXtomoScan:
            if self.extra_options.get("hdf5_entry", None) is not None:
                kwargs["entry"] = self.extra_options["hdf5_entry"]
            if self.extra_options.get("nx_version", None) is not None:
                kwargs["nx_version"] = self.extra_options["nx_version"]
        if self._scanner is EDFTomoScan:
            # Assume 1 frame per file (otherwise too long to open each file)
            kwargs["n_frames"] = 1

        self.dataset_scanner = self._scanner(  # pylint: disable=E1102
            self.location, ignore_projections=self._ignore_projections, **kwargs
        )

        if self._ignore_projections is not None:
            self.logger.info("Excluding projections: %s" % str(self._ignore_projections))

        if nabu_env_settings.skip_tomoscan_checks:
            self.logger.warning(
                " WARNING: according to nabu_env_settings.skip_tomoscan_checks, skipping virtual layout integrity check of tomoscan which is time consuming"
            )
            self.dataset_scanner.set_check_behavior(run_check=False, raise_error=False)

        self.raw_flats = self.dataset_scanner.flats
        self.raw_darks = self.dataset_scanner.darks
        self.n_angles = len(self.dataset_scanner.projections)
        self.radio_dims = (self.dataset_scanner.dim_1, self.dataset_scanner.dim_2)
        self._radio_dims_notbinned = self.radio_dims  # COMPAT

    def _finish_init(self):
        pass

    def _set_default_dataset_values(self):
        self._detector_tilt = None
        self.translations = None
        self.ctf_translations = None
        self.axis_position = None
        self._rotation_angles = None
        self.z_per_proj = None
        self.x_per_proj = None
        self._energy = None
        self._pixel_size = None
        self._distance = None
        self._flats_srcurrent = None
        self._projections = None
        self._projections_srcurrent = None
        self._reduced_flats = None
        self._reduced_darks = None

    @property
    def energy(self):
        """
        Return the energy in kev.
        """
        if self._energy is None:
            self._energy = self.dataset_scanner.energy
        return self._energy

    @energy.setter
    def energy(self, val):
        self._energy = val

    @property
    def distance(self):
        """
        Return the sample-detector distance in meters.
        """
        if self._distance is None:
            self._distance = abs(self.dataset_scanner.distance)
        return self._distance

    @distance.setter
    def distance(self, val):
        self._distance = val

    @property
    def pixel_size(self):
        """
        Return the pixel size in microns.
        """
        # TODO X and Y pixel size
        if self._pixel_size is None:
            self._pixel_size = self.dataset_scanner.pixel_size * 1e6
        return self._pixel_size

    @pixel_size.setter
    def pixel_size(self, val):
        self._pixel_size = val

    def _get_rotation_angles(self):
        return self._rotation_angles  # None by default

    @property
    def rotation_angles(self):
        """
        Return the rotation angles in radians.
        """
        return self._get_rotation_angles()

    @rotation_angles.setter
    def rotation_angles(self, angles):
        self._rotation_angles = angles

    def _is_halftomo(self):
        return None  # base class

    @property
    def is_halftomo(self):
        """
        Indicates whether the current dataset was performed with half acquisition.
        """
        return self._is_halftomo()

    @property
    def detector_tilt(self):
        """
        Return the detector tilt in degrees
        """
        return self._detector_tilt

    @detector_tilt.setter
    def detector_tilt(self, tilt):
        self._detector_tilt = tilt

    def _get_srcurrent(self, frame_type):
        # To be implemented by inheriting class
        return None

    @property
    def projections(self):
        if self._projections is None:
            self._projections = self.dataset_scanner.projections
        return self._projections

    @projections.setter
    def projections(self, val):
        raise ValueError

    @property
    def projections_srcurrent(self):
        """
        Return the synchrotron electric current for each projection.
        """
        if self._projections_srcurrent is None:
            self._projections_srcurrent = self._get_srcurrent("radios")  # pylint: disable=E1128
        return self._projections_srcurrent

    @projections_srcurrent.setter
    def projections_srcurrent(self, val):
        self._projections_srcurrent = val

    @property
    def flats_srcurrent(self):
        """
        Return the synchrotron electric current for each flat image.
        """
        if self._flats_srcurrent is None:
            self._flats_srcurrent = self._get_srcurrent("flats")  # pylint: disable=E1128
        return self._flats_srcurrent

    @flats_srcurrent.setter
    def flats_srcurrent(self, val):
        self._flats_srcurrent = val

    def check_defined_attribute(self, name, error_msg=None):
        """
        Utility function to check that a given attribute is defined.
        """
        if getattr(self, name, None) is None:
            raise ValueError(error_msg or str("No information on %s was found in the dataset" % name))

    @property
    def flats(self):
        """
        Return the REDUCED flat-field images. Either by reducing (median) the raw flats, or a user-defined reduced flats.
        """
        if self._reduced_flats is None:
            self._reduced_flats = self.get_reduced_flats()
        return self._reduced_flats

    @flats.setter
    def flats(self, val):
        self._reduced_flats = val

    @property
    def darks(self):
        """
        Return the REDUCED flat-field images. Either by reducing (mean) the raw darks, or a user-defined reduced darks.
        """
        if self._reduced_darks is None:
            self._reduced_darks = self.get_reduced_darks()
        return self._reduced_darks

    @darks.setter
    def darks(self, val):
        self._reduced_darks = val


class EDFDatasetAnalyzer(DatasetAnalyzer):
    """
    EDF Dataset analyzer for legacy ESRF acquisitions
    """

    _scanner = EDFTomoScan
    kind = "edf"

    def _finish_init(self):
        pass

    def _get_flats_darks(self):
        return

    @property
    def hdf5_entry(self):
        """
        Return the HDF5 entry of the current dataset.
        Not applicable for EDF (return None)
        """
        return None

    def _is_halftomo(self):
        return None

    def _get_rotation_angles(self):
        return np.deg2rad(self.dataset_scanner.rotation_angle())

    def get_reduced_flats(self, **reader_kwargs):
        if self.raw_flats in [None, {}]:
            raise FileNotFoundError("No reduced flat ('refHST') found in %s" % self.location)
        # A few notes:
        # (1) In principle we could do the reduction (mean/median) from raw frames (ref_xxxx_yyyy)
        #   but for legacy datasets it's always already done (by fasttomo3), and EDF support is supposed to be dropped on our side
        # (2) We use EDFStackReader class to handle the possible additional data modifications
        #  (eg. subsampling, binning, distortion correction...)
        # (3) The following spawns one reader instance per file, which is not elegant,
        #   but in principle there are typically 1-2 reduced flats in a scan
        readers = {k: EDFStackReader([self.raw_flats[k].file_path()], **reader_kwargs) for k in self.raw_flats.keys()}
        return {k: readers[k].load_data()[0] for k in self.raw_flats.keys()}

    def get_reduced_darks(self, **reader_kwargs):
        # See notes in get_reduced_flats() above
        if self.raw_darks in [None, {}]:
            raise FileNotFoundError("No reduced dark ('darkend.edf' or 'dark.edf') found in %s" % self.location)
        readers = {k: EDFStackReader([self.raw_darks[k].file_path()], **reader_kwargs) for k in self.raw_darks.keys()}
        return {k: readers[k].load_data()[0] for k in self.raw_darks.keys()}

    @property
    def files(self):
        return sorted([u.file_path() for u in self.dataset_scanner.projections.values()])

    def get_reader(self, **kwargs):
        return EDFStackReader(self.files, **kwargs)


class HDF5DatasetAnalyzer(DatasetAnalyzer):
    """
    HDF5 dataset analyzer
    """

    _scanner = NXtomoScan
    kind = "nx"
    # We could import the 1000+ LoC  nxtomo.nxobject.nxdetector.ImageKey... or we can do this
    _image_key_value = {"flats": 1, "darks": 2, "radios": 0}
    #

    @property
    def z_translation(self):
        raw_data = np.array(self.dataset_scanner.z_translation)
        projs_idx = np.array(list(self.projections.keys()))
        filtered_data = raw_data[projs_idx]
        return 1.0e6 * filtered_data / self.pixel_size

    @property
    def x_translation(self):
        raw_data = np.array(self.dataset_scanner.x_translation)
        projs_idx = np.array(list(self.projections.keys()))
        filtered_data = raw_data[projs_idx]
        return 1.0e6 * filtered_data / self.pixel_size

    def _get_rotation_angles(self):
        if self._rotation_angles is None:
            angles = np.array(self.dataset_scanner.rotation_angle)
            projs_idx = np.array(list(self.projections.keys()))
            angles = angles[projs_idx]
            self._rotation_angles = np.deg2rad(angles)
        return self._rotation_angles

    def _get_dataset_hdf5_url(self):
        if len(self.projections) > 0:
            frames_to_take = self.projections
        elif len(self.raw_flats) > 0:
            frames_to_take = self.raw_flats
        elif len(self.raw_darks) > 0:
            frames_to_take = self.raw_darks
        else:
            raise ValueError("No projections, no flats and no darks ?!")
        first_proj_idx = sorted(frames_to_take.keys())[0]
        first_proj_url = frames_to_take[first_proj_idx]
        return DataUrl(
            file_path=first_proj_url.file_path(), data_path=first_proj_url.data_path(), data_slice=None, scheme="silx"
        )

    @property
    def dataset_hdf5_url(self):
        return self._get_dataset_hdf5_url()

    @property
    def hdf5_entry(self):
        """
        Return the HDF5 entry of the current dataset
        """
        return self.dataset_scanner.entry

    def _is_halftomo(self):
        try:
            is_halftomo = self.dataset_scanner.field_of_view.value.lower() == "half"
        except:
            is_halftomo = None
        return is_halftomo

    def get_data_slices(self, what):
        """
        Return indices in the data volume where images correspond to a given kind.

        Parameters
        ----------
        what: str
            Which keys to get. Can be "projections", "flats", "darks"

        Returns
        --------
        slices: list of slice
            A list where each item is a slice.
        """
        name_to_attr = {
            "projections": self.projections,
            "flats": self.raw_flats,
            "darks": self.raw_darks,
        }
        check_supported(what, name_to_attr.keys(), "image type")
        images = name_to_attr[what]  # dict
        # we can't directly use set() on slice() object (unhashable). Use tuples
        slices = set()
        for du in get_compacted_dataslices(images).values():
            if du.data_slice() is not None:
                s = (du.data_slice().start, du.data_slice().stop)
            else:
                s = None
            slices.add(s)
        slices_list = [slice(item[0], item[1]) if item is not None else None for item in list(slices)]
        return slices_list

    def _select_according_to_frame_type(self, data, frame_type):
        if data is None:
            return None
        return data[self.dataset_scanner.image_key_control == self._image_key_value[frame_type]]

    def get_reduced_flats(self, method="median", force_reload=False, **reader_kwargs):
        dkrf_reader = NXDarksFlats(
            self.dataset_hdf5_url.file_path(), data_path=self.dataset_hdf5_url.data_path(), **reader_kwargs
        )
        return dkrf_reader.get_reduced_flats(method=method, force_reload=force_reload, as_dict=True)

    def get_reduced_darks(self, method="mean", force_reload=False, **reader_kwargs):
        dkrf_reader = NXDarksFlats(
            self.dataset_hdf5_url.file_path(), data_path=self.dataset_hdf5_url.data_path(), **reader_kwargs
        )
        return dkrf_reader.get_reduced_darks(method=method, force_reload=force_reload, as_dict=True)

    def _get_srcurrent(self, frame_type):
        return self._select_according_to_frame_type(self.dataset_scanner.electric_current, frame_type)

    def frames_slices(self, frame_type):
        """
        Return a list of slice objects corresponding to the data corresponding to "frame_type".
        For example, if the dataset flats are located at indices [1, 2, ..., 99], then
        frame_slices("flats") will return [slice(0, 100)].
        """
        return indices_to_slices(
            np.where(self.dataset_scanner.image_key_control == self._image_key_value[frame_type])[0]
        )

    def get_reader(self, **kwargs):
        return NXTomoReader(self.dataset_hdf5_url.file_path(), data_path=self.dataset_hdf5_url.data_path(), **kwargs)


def analyze_dataset(dataset_path, extra_options=None, logger=None):
    if not (os.path.isdir(dataset_path)):
        if not (os.path.isfile(dataset_path)):
            raise ValueError("Error: %s no such file or directory" % dataset_path)
        if not (is_hdf5_extension(os.path.splitext(dataset_path)[-1].replace(".", ""))):
            raise ValueError("Error: expected a HDF5 file")
        dataset_analyzer_class = HDF5DatasetAnalyzer
    else:  # directory -> assuming EDF
        dataset_analyzer_class = EDFDatasetAnalyzer
    dataset_structure = dataset_analyzer_class(dataset_path, extra_options=extra_options, logger=logger)
    return dataset_structure


def get_radio_pair(dataset_info, radio_angles: tuple, return_indices=False):
    """
    Get closest radios at radio_angles[0] and radio_angles[1]
    angles must be in angles

    Parameters
    ----------
    dataset_info: `DatasetAnalyzer` instance
        Data structure with the dataset information
    radio_angles: tuple
        tuple of two elements: angles (in radian) to get
    return_indices: bool, optional
        Whether to return radios indices along with the radios array.

    Returns
    -------
    res: array or tuple
        If return_indices is True, return a tuple (radios, indices).
        Otherwise, return an array with the radios.
    """
    if not (isinstance(radio_angles, tuple) and len(radio_angles) == 2):
        raise TypeError("radio_angles should be a tuple of two elements.")
    if not isinstance(radio_angles[0], (np.floating, float)) or not isinstance(radio_angles[1], (np.floating, float)):
        raise TypeError(
            f"radio_angles should be float. Get {type(radio_angles[0])} and {type(radio_angles[1])} instead"
        )

    radios_indices = []
    radios_indices = sorted(dataset_info.projections.keys())
    angles = dataset_info.rotation_angles
    angles = angles - angles.min()
    i_radio_1 = np.argmin(np.abs(angles - radio_angles[0]))
    i_radio_2 = np.argmin(np.abs(angles - radio_angles[1]))
    radios_indices = [radios_indices[i_radio_1], radios_indices[i_radio_2]]
    n_radios = 2
    radios = np.zeros((n_radios,) + dataset_info.radio_dims[::-1], "f")
    for i in range(n_radios):
        radio_idx = radios_indices[i]
        radios[i] = get_data(dataset_info.projections[radio_idx]).astype("f")
    if return_indices:
        return radios, radios_indices
    else:
        return radios



from itertools import product
import numpy as np
import pytest
from nabu.processing.unsharp import UnsharpMask
from nabu.processing.unsharp_opencl import OpenclUnsharpMask, __have_opencl__ as __has_pyopencl__
from nabu.cuda.utils import __has_pycuda__, get_cuda_context
from nabu.testutils import get_data

if __has_pyopencl__:
    from pyopencl import CommandQueue
    import pyopencl.array as parray
    from silx.opencl.common import ocl
if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.processing.unsharp_cuda import CudaUnsharpMask

try:
    from skimage.filters import unsharp_mask

    __has_skimage__ = True
except ImportError:
    __has_skimage__ = False


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("brain_phantom.npz")["data"]
    cls.imagej_results = get_data("dirac_unsharp_imagej.npz")
    cls.tol = 1e-4
    cls.sigma = 1.6
    cls.coeff = 0.5
    if __has_pycuda__:
        cls.ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.cl_ctx = ocl.create_context()
    yield
    if __has_pycuda__:
        cls.ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestUnsharp:
    def get_reference_result(self, method, data=None):
        if data is None:
            data = self.data
        unsharp_mask = UnsharpMask(data.shape, self.sigma, self.coeff, method=method)
        return unsharp_mask.unsharp(data)

    def check_result(self, result, method, data=None, error_msg_prefix=None):
        reference = self.get_reference_result(method, data=data)
        mae = np.max(np.abs(result - reference))
        err_msg = str(
            "%s: max error is too high with method=%s: %.2e > %.2e" % (error_msg_prefix or "", method, mae, self.tol)
        )
        assert mae < self.tol, err_msg

    @pytest.mark.skipif(not (__has_skimage__), reason="Need scikit-image for this test")
    def test_mode_gaussian(self):
        dirac = np.zeros((43, 43), "f")
        dirac[dirac.shape[0] // 2, dirac.shape[1] // 2] = 1
        sigma_list = [0.2, 0.5, 1.0, 2.0, 3.0]
        coeff_list = [0.5, 1.0, 3.0]
        for sigma, coeff in product(sigma_list, coeff_list):
            res = UnsharpMask(dirac.shape, sigma, coeff, method="gaussian").unsharp(dirac)
            ref = unsharp_mask(dirac, radius=sigma, amount=coeff, preserve_range=True)
            assert np.max(np.abs(res - ref)) < 1e-6, "Something wrong with mode='gaussian', sigma=%.2f, coeff=%.2f" % (
                sigma,
                coeff,
            )

    def test_mode_imagej(self):
        dirac = np.zeros(self.imagej_results["images"][0].shape, dtype="f")
        dirac[dirac.shape[0] // 2, dirac.shape[1] // 2] = 1
        for sigma, coeff, ref in zip(
            self.imagej_results["sigma"], self.imagej_results["amount"], self.imagej_results["images"]
        ):
            res = UnsharpMask(dirac.shape, sigma, coeff, method="imagej").unsharp(dirac)
            assert np.max(np.abs(res - ref)) < 1e-3, "Something wrong with mode='imagej', sigma=%.2f, coeff=%.2f" % (
                sigma,
                coeff,
            )

    @pytest.mark.skipif(not (__has_pyopencl__), reason="Need pyopencl for this test")
    def test_opencl_unsharp(self):
        cl_queue = CommandQueue(self.cl_ctx)
        d_image = parray.to_device(cl_queue, self.data)
        d_out = parray.zeros_like(d_image)
        for method in OpenclUnsharpMask.avail_methods:
            d_image = parray.to_device(cl_queue, self.data)
            d_out = parray.zeros_like(d_image)

            opencl_unsharp = OpenclUnsharpMask(self.data.shape, self.sigma, self.coeff, method=method, ctx=self.cl_ctx)
            opencl_unsharp.unsharp(d_image, output=d_out)
            res = d_out.get()
            self.check_result(res, method, error_msg_prefix="OpenclUnsharpMask")

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_unsharp(self):
        d_image = garray.to_gpu(self.data)
        d_out = garray.zeros_like(d_image)
        for method in CudaUnsharpMask.avail_methods:
            cuda_unsharp = CudaUnsharpMask(self.data.shape, self.sigma, self.coeff, method=method, ctx=self.ctx)
            cuda_unsharp.unsharp(d_image, output=d_out)
            res = d_out.get()
            self.check_result(res, method, error_msg_prefix="CudaUnsharpMask")



from nabu.stitching.utils.utils import has_itk, find_shift_with_itk
from scipy.ndimage import shift as shift_scipy
import numpy
import pytest
from nabu.testutils import get_data


@pytest.mark.parametrize("data_type", (numpy.float32, numpy.uint16))
@pytest.mark.skipif(not has_itk, reason="itk not installed")
def test_find_shift_with_itk(data_type):
    shift = (5, 2)
    img1 = get_data("chelsea.npz")["data"].astype(data_type)
    img2 = shift_scipy(
        img1.copy(),
        shift=shift,
        order=1,
    )

    img1 = img1[10:-10, 10:-10]
    img2 = img2[10:-10, 10:-10]
    assert find_shift_with_itk(img1=img1, img2=img2) == shift



from tomoscan.io import HDF5File
from .. import version
from ..utils import list_match_queries
from ..pipeline.config import parse_nabu_config_file
from ..pipeline.config_validators import convert_to_int
from .cli_configs import ReconstructConfig
from .utils import parse_params_values


def update_reconstruction_start_end(conf_dict, user_indices):
    if len(user_indices) == 0:
        return
    rec_cfg = conf_dict["reconstruction"]
    err = None

    val_int, conv_err = convert_to_int(user_indices)
    if conv_err is None:
        start_z = user_indices
        end_z = user_indices
    else:
        if user_indices in ["first", "middle", "last"]:
            start_z = user_indices
            end_z = user_indices
        elif user_indices == "all":
            start_z = 0
            end_z = -1
        elif "-" in user_indices:
            try:
                start_z, end_z = user_indices.split("-")
                start_z = int(start_z)
                end_z = int(end_z)
            except Exception as exc:
                err = "Could not interpret slice indices '%s': %s" % (user_indices, str(exc))
        else:
            err = "Could not interpret slice indices: %s" % user_indices
        if err is not None:
            print(err)
            exit(1)
    rec_cfg["start_z"] = start_z
    rec_cfg["end_z"] = end_z


def get_log_file(arg_logfile, legacy_arg_logfile, forbidden=None):
    default_arg_val = ""
    # Compat. log_file --> logfile
    if legacy_arg_logfile != default_arg_val:
        logfile = legacy_arg_logfile
    else:
        logfile = arg_logfile
    #
    if forbidden is None:
        forbidden = []
    for forbidden_val in forbidden:
        if logfile == forbidden_val:
            print("Error: --logfile argument cannot have the value %s" % forbidden_val)
            exit(1)
    if logfile == "":
        logfile = True
    return logfile


def get_reconstructor(args, overwrite_options=None):
    # Imports are done here, otherwise "nabu --version" takes forever
    from ..pipeline.fullfield.processconfig import ProcessConfig
    from ..pipeline.fullfield.reconstruction import FullFieldReconstructor

    #

    logfile = get_log_file(args["logfile"], args["log_file"], forbidden=[args["input_file"]])
    conf_dict = parse_nabu_config_file(args["input_file"])
    update_reconstruction_start_end(conf_dict, args["slice"].strip())
    if overwrite_options is not None:
        for option_key, option_val in overwrite_options.items():
            opt_section, opt_name = option_key.split("/")
            conf_dict[opt_section][opt_name] = option_val

    proc = ProcessConfig(conf_dict=conf_dict, create_logger=logfile)
    logger = proc.logger

    logger.info("Going to reconstruct slices (%d, %d)" % (proc.rec_region["start_z"], proc.rec_region["end_z"]))

    # Get extra options
    extra_options = {
        "gpu_mem_fraction": args["gpu_mem_fraction"],
        "cpu_mem_fraction": args["cpu_mem_fraction"],
        "chunk_size": args["max_chunk_size"] if args["max_chunk_size"] > 0 else None,
        "margin": args["phase_margin"],
        "force_grouped_mode": bool(args["force_use_grouped_pipeline"]),
    }

    reconstructor = FullFieldReconstructor(proc, logger=logger, extra_options=extra_options)
    return reconstructor


def list_hdf5_entries(fname):
    with HDF5File(fname, "r") as f:
        entries = list(f.keys())
    return entries


def main():
    args = parse_params_values(
        ReconstructConfig,
        parser_description=f"Perform a tomographic reconstruction.",
        program_version="nabu " + version,
    )
    # Get extra options
    extra_options = {
        "gpu_mem_fraction": args["gpu_mem_fraction"],
        "cpu_mem_fraction": args["cpu_mem_fraction"],
        "chunk_size": args["max_chunk_size"] if args["max_chunk_size"] > 0 else None,
        "margin": args["phase_margin"],
        "force_grouped_mode": bool(args["force_use_grouped_pipeline"]),
    }
    #

    logfile = get_log_file(args["logfile"], args["log_file"], forbidden=[args["input_file"]])
    conf_dict = parse_nabu_config_file(args["input_file"])
    update_reconstruction_start_end(conf_dict, args["slice"].strip())

    # Imports are done here, otherwise "nabu --version" takes forever
    from ..pipeline.fullfield.processconfig import ProcessConfig
    from ..pipeline.fullfield.reconstruction import FullFieldReconstructor

    #

    hdf5_entries = conf_dict["dataset"].get("hdf5_entry", "").strip(",")
    # spit by coma and remove empty spaces
    hdf5_entries = [e.strip() for e in hdf5_entries.split(",")]
    # clear '/' at beginning of the entry (so both entry like 'entry0000' and '/entry0000' are handled)
    hdf5_entries = [e.lstrip("/") for e in hdf5_entries]

    if hdf5_entries != [""]:
        file_hdf5_entries = list_hdf5_entries(conf_dict["dataset"]["location"])
        hdf5_entries = list_match_queries(file_hdf5_entries, hdf5_entries)
        if hdf5_entries == []:
            raise ValueError("No entry found matching pattern '%s'" % conf_dict["dataset"]["hdf5_entry"])

    for hdf5_entry in hdf5_entries:
        if len(hdf5_entries) > 1:
            print("-" * 80)
            print("Processing entry: %s" % hdf5_entry)
            print("-" * 80)
        conf_dict["dataset"]["hdf5_entry"] = hdf5_entry
        proc = ProcessConfig(conf_dict=conf_dict, create_logger=logfile)  # logger is in append mode
        logger = proc.logger
        logger.info("Going to reconstruct slices (%d, %d)" % (proc.rec_region["start_z"], proc.rec_region["end_z"]))

        R = FullFieldReconstructor(proc, logger=logger, extra_options=extra_options)
        proc = R.process_config

        R.reconstruct()
        R.finalize_files_saving()
    return 0


if __name__ == "__main__":
    main()



import numpy as np
from scipy.signal import convolve2d
from scipy.fft import rfft

from ..utils import deprecation_warning, is_scalar
from ..resources.logger import LoggerOrPrint

try:
    from algotom.prep.calculation import find_center_vo, find_center_360

    __have_algotom__ = True
except ImportError:
    __have_algotom__ = False


class SinoCor:
    """
    This class has 2 methods:
        - overlap. Find a rough estimate of COR
        - accurate. Try to refine COR to 1/10 pixel
    """

    def __init__(self, img_1, img_2, logger=None):
        """ """
        self.logger = LoggerOrPrint(logger)
        self.sx = img_1.shape[1]

        # algorithm cannot accept odd number of projs. This is handled in the SinoCORFinder class.
        nproj2 = img_1.shape[0]

        # extract upper and lower part of sinogram, flipping H the upper part
        self.data1 = img_1
        self.data2 = img_2

        self.rcor_abs = round(self.sx / 2.0)
        self.cor_acc = round(self.sx / 2.0)

        # parameters for overlap sino - rough estimation

        # default sliding ROI is 20% of the width of the detector
        # the maximum size of ROI in the "right" case is 2*(self.sx - COR)
        # ex: 2048 pixels, COR= 2000, window_width should not exceed 96!

        self.window_width = round(self.sx / 5)

    @staticmethod
    def schift(mat, val):
        ker = np.zeros((3, 3))
        s = 1.0
        if val < 0:
            s = -1.0
        val = s * val
        ker[1, 1] = 1 - val
        if s > 0:
            ker[1, 2] = val
        else:
            ker[1, 0] = val
        mat = convolve2d(mat, ker, mode="same")
        return mat

    def overlap(self, side="right", window_width=None):
        """
        Compute COR by minimizing difference of circulating ROI

         - side:         preliminary knowledge if the COR is on right or left
         - window_width: width of ROI that will slide on the other part of the sinogram
                         by default, 20% of the width of the detector.
        """

        if window_width is None:
            window_width = self.window_width

        if not (window_width & 1):
            window_width -= 1

        # number of pixels where the window will "slide".
        n = self.sx - int(window_width)
        nr = range(n)

        dmax = 1000000000.0
        imax = 0

        # Should we do both right and left and take the minimum "diff" of the 2 ?
        # windows self.data2 moves over self.data1, measure the width of the histogram and retains the smaller one.
        if side == "right":
            for i in nr:
                imout = self.data1[:, n - i : n - i + window_width] - self.data2[:, 0:window_width]
                diff = imout.max() - imout.min()
                if diff < dmax:
                    dmax = diff
                    imax = i
            self.cor_abs = self.sx - (imax + window_width + 1.0) / 2.0
            self.cor_rel = self.sx / 2 - (imax + window_width + 1.0) / 2.0
        elif side == "left":
            for i in nr:
                imout = self.data1[:, i : i + window_width] - self.data2[:, self.sx - window_width : self.sx]
                diff = imout.max() - imout.min()
                if diff < dmax:
                    dmax = diff
                    imax = i
            self.cor_abs = (imax + window_width - 1.0) / 2
            self.cor_rel = self.cor_abs - self.sx / 2.0 - 1
        else:
            raise ValueError(f"Invalid side given ({side}). should be 'left' or 'right'")
        if imax < 1:
            self.logger.warning("sliding width %d seems too large!" % window_width)
        self.rcor_abs = round(self.cor_abs)
        return self.rcor_abs

    def accurate(self, neighborhood=7, shift_value=0.1):
        """
        refine the calculation around COR integer pre-calculated value
        The search will be executed in the defined neighborhood

        Parameters
        -----------
        neighborhood: int
            Parameter for accurate calculation in the vicinity of the rough estimate.
            It must be an odd number.
            0.1 pixels float shifts will be performed over this number of pixel
        """
        # define the H-size (odd) of the window one can use to find the best overlap moving finely over ng pixels
        if not (neighborhood & 1):
            neighborhood += 1
        ng2 = int(neighborhood / 2)

        # pleft and pright are the number of pixels available on the left and the right of the cor position
        # to slide a window
        pleft = self.rcor_abs - ng2
        pright = self.sx - self.rcor_abs - ng2 - 1

        # the maximum window to slide is restricted by the smaller side
        if pleft > pright:
            p_sign = 1
            xwin = 2 * (self.sx - self.rcor_abs - ng2) - 1
        else:
            p_sign = -1
            xwin = 2 * (self.rcor_abs - ng2) + 1

        # Note that xwin is odd
        xc1 = self.rcor_abs - int(xwin / 2)
        xc2 = self.sx - self.rcor_abs - int(xwin / 2) - 1

        im1 = self.data1[:, xc1 : xc1 + xwin]
        im2 = self.data2[:, xc2 : xc2 + xwin]

        pixs = p_sign * (np.arange(neighborhood) - ng2)
        diff0 = 1000000000.0

        isfr = shift_value * np.arange(10)
        self.cor_acc = self.rcor_abs

        for pix in pixs:
            x0 = xc1 + pix
            for isf in isfr:
                if isf != 0:
                    ims = self.schift(self.data1[:, x0 : x0 + xwin].copy(), -p_sign * isf)
                else:
                    ims = self.data1[:, x0 : x0 + xwin]

                imout = ims - self.data2[:, xc2 : xc2 + xwin]
                diff = imout.max() - imout.min()
                if diff < diff0:
                    self.cor_acc = self.rcor_abs + (pix + p_sign * isf) / 2.0
                    diff0 = diff
        return self.cor_acc

    # Aliases
    estimate_cor_coarse = overlap
    estimate_cor_fine = accurate


class SinoCorInterface:
    """
    A class that mimics the interface of CenterOfRotation, while calling SinoCor
    """

    def __init__(self, logger=None, **kwargs):
        self._logger = logger

    def find_shift(
        self,
        img_1,
        img_2,
        side="right",
        window_width=None,
        neighborhood=7,
        shift_value=0.1,
        return_relative_to_middle=None,
        **kwargs,
    ):

        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationCoarseToFine.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---

        cor_finder = SinoCor(img_1, img_2, logger=self._logger)
        cor_finder.estimate_cor_coarse(side=side, window_width=window_width)
        cor = cor_finder.estimate_cor_fine(neighborhood=neighborhood, shift_value=shift_value)
        # offset will be added later - keep compatibility with result from AlignmentBase.find_shift()
        if return_relative_to_middle:
            return cor - (img_1.shape[1] - 1) / 2
        else:
            return cor


class CenterOfRotationFourierAngles:
    """This CoR estimation algo is proposed by V. Valls (BCU). It is based on the Fourier
    transform of the columns on the sinogram.
    It requires an initial guesss of the CoR wich is retrieved from
    dataset_info.dataset_scanner.x_rotation_axis_pixel_position. It is assumed in mm and pixel size in um.
    Options are (for the moment) hard-coded in the SinoCORFinder.cor_finder.extra_options dict.
    """

    def __init__(self, *args, **kwargs):
        pass

    def _convert_from_fft_2_fftpack_format(self, f_signal, o_signal_length):
        """
        Converts a scipy.fft.rfft into the (legacy) scipy.fftpack.rfft format.
        The fftpack.rfft returns a (roughly) twice as long array as fft.rfft as the latter returns an array
        of complex numbers wheras the former returns an array with real and imag parts in consecutive
        spots in the array.

        Parameters
        ----------
        f_signal : array_like
            The output of scipy.fft.rfft(signal)
        o_signal_length : int
            Size of the original signal (before FT).

        Returns
        -------
        out
            The rfft converted to the fftpack.rfft format (roughly twice as long).
        """
        out = np.zeros(o_signal_length, dtype=np.float32)
        if o_signal_length % 2 == 0:
            out[0] = f_signal[0].real
            out[1::2] = f_signal[1:].real
            out[2::2] = f_signal[1:-1].imag
        else:
            out[0] = f_signal[0].real
            out[1::2] = f_signal[1:].real
            out[2::2] = f_signal[1:].imag
        return out

    def _freq_radio(self, sinos, ifrom, ito):
        size = (sinos.shape[0] + sinos.shape[0] % 2) // 2
        fs = np.empty((size, sinos.shape[1]))
        for i in range(ifrom, ito):
            line = sinos[:, i]
            f_signal = rfft(line)
            f_signal = self._convert_from_fft_2_fftpack_format(f_signal, line.shape[0])
            f = np.abs(f_signal[: (f_signal.size - 1) // 2 + 1])
            f2 = np.abs(f_signal[(f_signal.size - 1) // 2 + 1 :][::-1])
            if len(f) > len(f2):
                f[1:] += f2
            else:
                f[0:] += f2
            fs[:, i] = f
        with np.errstate(divide="ignore", invalid="ignore", under="ignore"):
            fs = np.log(fs)
        return fs

    def gaussian(self, p, x):
        return p[3] + p[2] * np.exp(-((x - p[0]) ** 2) / (2 * p[1] ** 2))

    def tukey(self, p, x):
        pos, std, alpha, height, background = p
        alpha = np.clip(alpha, 0, 1)
        pi = np.pi
        inv_alpha = 1 - alpha
        width = std / (1 - alpha * 0.5)
        xx = (np.abs(x - pos) - (width * 0.5 * inv_alpha)) / (width * 0.5 * alpha)
        xx = np.clip(xx, 0, 1)
        return (0.5 + np.cos(pi * xx) * 0.5) * height + background

    def sinlet(self, p, x):
        std = p[1] * 2.5
        lin = np.maximum(0, std - np.abs(p[0] - x)) * 0.5 * np.pi / std
        return p[3] + p[2] * np.sin(lin)

    def _px(self, detector_width, abs_pos, near_width, near_std, crop_around_cor, near_step):
        sym_range = None
        if abs_pos is not None:
            if crop_around_cor:
                sym_range = int(abs_pos - near_std * 2), int(abs_pos + near_std * 2)

        window = near_width
        if sym_range is not None:
            xx_from = max(window, sym_range[0])
            xx_to = max(xx_from, min(detector_width - window, sym_range[1]))
            if xx_from == xx_to:
                sym_range = None
        if sym_range is None:
            xx_from = window
            xx_to = detector_width - window

        xx = np.arange(xx_from, xx_to, near_step)

        return xx

    def _symmetry_correlation(self, px, array, angles, window, shift_sino):
        if shift_sino:
            shift_index = np.argmin(np.abs(angles - np.pi)) - np.argmin(np.abs(angles - 0))
        else:
            shift_index = None
        px_from = int(px[0])
        px_to = int(np.ceil(px[-1]))
        f_coef = np.empty(len(px))
        f_array = self._freq_radio(array, px_from - window, px_to + window)
        if shift_index is not None:
            shift_array = np.empty(array.shape, dtype=array.dtype)
            shift_array[0 : len(shift_array) - shift_index, :] = array[shift_index:, :]
            shift_array[len(shift_array) - shift_index :, :] = array[:shift_index, :]
            f_shift_array = self._freq_radio(shift_array, px_from - window, px_to + window)
        else:
            f_shift_array = f_array

        for j, x in enumerate(px):
            i = int(np.floor(x))
            if x - i > 0.4:  # TO DO : Specific to near_step = 0.5?
                f_left = f_array[:, i - window : i]
                f_right = f_shift_array[:, i + 1 : i + window + 1][:, ::-1]
            else:
                f_left = f_array[:, i - window : i]
                f_right = f_shift_array[:, i : i + window][:, ::-1]
            with np.errstate(divide="ignore", invalid="ignore"):
                f_coef[j] = np.sum(np.abs(f_left - f_right))
        return f_coef

    def _cor_correlation(self, px, abs_pos, near_std, signal, near_weight, near_alpha):
        if abs_pos is not None:
            if signal == "sinlet":
                coef = self.sinlet((abs_pos, near_std, -near_weight, 1), px)
            elif signal == "gaussian":
                coef = self.gaussian((abs_pos, near_std, -near_weight, 1), px)
            elif signal == "tukey":
                coef = self.tukey((abs_pos, near_std * 2, near_alpha, -near_weight, 1), px)
            else:
                raise ValueError("Shape unsupported")
        else:
            coef = np.ones_like(px)
        return coef

    def find_shift(
        self,
        sino,
        angles=None,
        side="center",
        near_std=100,
        near_width=20,
        shift_sino=True,
        crop_around_cor=False,
        signal="tukey",
        near_weight=0.1,
        near_alpha=0.5,
        near_step=0.5,
        return_relative_to_middle=None,
    ):
        detector_width = sino.shape[1]

        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationFourierAngles.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---

        if angles is None:
            angles = np.linspace(0, 2 * np.pi, sino.shape[0], endpoint=True)
        increment = np.abs(angles[0] - angles[1])
        if np.abs(angles[0] - angles[-1]) < (360 - 0.5) * np.pi / 180 - increment:
            raise ValueError("Not enough angles, estimator skipped")

        if is_scalar(side):
            abs_pos = side
        # COMPAT.
        elif side == "near":
            deprecation_warning(
                "side='near' is deprecated, please use side=<a scalar>", do_print=True, func_name="fourier_angles_near"
            )
            abs_pos = detector_width // 2
        ##.
        elif side == "center":
            abs_pos = detector_width // 2
        elif side == "left":
            abs_pos = detector_width // 4
        elif side == "right":
            abs_pos = detector_width * 3 // 4
        else:
            raise ValueError(f"side '{side}' is not handled")

        px = self._px(detector_width, abs_pos, near_width, near_std, crop_around_cor, near_step)

        coef_f = self._symmetry_correlation(
            px,
            sino,
            angles,
            near_width,
            shift_sino,
        )
        coef_p = self._cor_correlation(px, abs_pos, near_std, signal, near_weight, near_alpha)
        coef = coef_f * coef_p

        if len(px) > 0:
            cor = px[np.argmin(coef)] - (detector_width - 1) / 2
        else:
            # raise ValueError ?
            cor = None
        if not (return_relative_to_middle):
            cor += (detector_width - 1) / 2
        return cor

    __call__ = find_shift


class CenterOfRotationVo:
    """
    A wrapper around algotom 'find_center_vo' and 'find_center_360'.

    Nghia T. Vo, Michael Drakopoulos, Robert C. Atwood, and Christina Reinhard,
    "Reliable method for calculating the center of rotation in parallel-beam tomography,"
    Opt. Express 22, 19078-19086 (2014)
    """

    default_extra_options = {}

    def __init__(self, logger=None, verbose=False, extra_options=None):
        if not (__have_algotom__):
            raise ImportError("Need the 'algotom' package")
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def find_shift(
        self,
        sino,
        halftomo=False,
        is_360=False,
        win_width=100,
        side="center",
        search_width_fraction=0.1,
        step=0.25,
        radius=4,
        ratio=0.5,
        dsp=True,
        ncore=None,
        hor_drop=None,
        ver_drop=None,
        denoise=True,
        norm=True,
        use_overlap=False,
        return_relative_to_middle=None,
    ):
        # COMPAT.
        if return_relative_to_middle is None:
            deprecation_warning(
                "The current default behavior is to return the shift relative the the middle of the image. In a future release, this function will return the shift relative to the left-most pixel. To keep the current behavior, please use 'return_relative_to_middle=True'.",
                do_print=True,
                func_name="CenterOfRotationVo.find_shift",
            )
            return_relative_to_middle = True  # the kwarg above will be False by default in a future release
        # ---

        if halftomo:
            side_algotom = {"left": 0, "right": 1}.get(side, None)
            cor, _, _, _ = find_center_360(
                sino, win_width, side=side_algotom, denoise=denoise, norm=norm, use_overlap=use_overlap, ncore=ncore
            )
        else:
            if is_360 and not (halftomo):
                # Take only one part of the sinogram and use "find_center_vo" - this works better in this case
                sino = sino[: sino.shape[0] // 2]

            sino_width = sino.shape[-1]
            search_width = int(search_width_fraction * sino_width)

            if side == "left":
                start, stop = 0, search_width
            elif side == "center":
                start, stop = sino_width // 2 - search_width, sino_width // 2 + search_width
            elif side == "right":
                start, stop = sino_width - search_width, sino_width
            elif is_scalar(side):
                # side is passed as an offset from the middle of detector
                side = side + (sino.shape[-1] - 1) / 2.0
                start, stop = max(0, side - search_width), min(sino_width, side + search_width)
            else:
                raise ValueError("Expected 'side' to be 'left', 'center', 'right' or a scalar")

            cor = find_center_vo(
                sino,
                start=start,
                stop=stop,
                step=step,
                radius=radius,
                ratio=ratio,
                dsp=dsp,
                ncore=ncore,
                hor_drop=hor_drop,
                ver_drop=ver_drop,
            )
        return cor if not (return_relative_to_middle) else cor - (sino.shape[1] - 1) / 2



from .double_flatfield import (
    DoubleFlatField,
    DoubleFlatField,
    check_shape,
    get_2D_3D_shape,
)
from ..misc.binning import get_binning_function


class DoubleFlatFieldVariableRegion(DoubleFlatField):
    def __init__(
        self,
        shape,
        result_url=None,
        binning_x=None,
        binning_z=None,
        detector_corrector=None,
    ):
        """This class provides the division by the double flat field.
        At variance with the standard class, it store as member the
        whole field, and performs the division by the proper region
        according to the positionings of the processed radios
        which is passed by the argument array sub_regions_per_radio to
        the method apply_double_flatfield_for_sub_regions
        """

        self.radios_shape = get_2D_3D_shape(shape)
        self.n_angles = self.radios_shape[0]
        self.shape = self.radios_shape[1:]
        self._init_filedump(result_url, None, detector_corrector)

        data = self._load_dff_full_dump()

        if (binning_z, binning_x) != (1, 1):
            print(" (binning_z, binning_x) ", (binning_z, binning_x))
            binning_function = get_binning_function((binning_z, binning_x))
            if binning_function is None:
                raise NotImplementedError(f"Binning factor for {(binning_z, binning_x)} is not implemented yet")

            self.data = binning_function(data)

        else:
            self.data = data

    def _load_dff_full_dump(self):
        res = self.reader.get_data(self.result_url)
        if self.detector_corrector is not None:
            self.detector_corrector.set_full_transformation()
            res = self.detector_corrector.transform(res, do_full=True)

        return res

    def apply_double_flatfield_for_sub_regions(self, radios, sub_regions_per_radio):
        """
        Apply the "double flatfield" filter on a chunk of radios.
        The processing is done in-place !
        """
        my_double_ff = self.data

        for i in range(radios.shape[0]):
            s_x, e_x, s_y, e_y = sub_regions_per_radio[i]

            dff = my_double_ff[s_y:e_y, s_x:e_x]

            check_shape(radios[i].shape, dff.shape, "radios")

            radios[i] /= dff
        return radios



import numpy as np
from ..utils import get_opencl_srcfile
from ..opencl.processing import OpenCLProcessing
from ..processing.padding_opencl import OpenCLPadding
from ..opencl.memcpy import OpenCLMemcpy2D
from .filtering import SinoFilter

try:
    from pyvkfft.opencl import VkFFTApp as clfft  # pylint: disable=E0401

    __has_vkfft__ = True
except:
    __has_vkfft__ = False


class OpenCLSinoFilter(SinoFilter):
    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        extra_options=None,
        opencl_options=None,
    ):
        self._opencl_options = opencl_options or {}
        self.opencl = OpenCLProcessing(**self._opencl_options)
        self.queue = self.opencl.queue
        super().__init__(sino_shape, filter_name=filter_name, padding_mode=padding_mode, extra_options=extra_options)
        self._init_kernels()

    def _init_fft(self):
        if not (__has_vkfft__):
            raise ImportError("Please install pyvkfft to use this class")
        self.fft = clfft(self.sino_padded_shape, np.float32, self.queue, r2c=True, ndim=1, inplace=False)

    def _allocate_memory(self):
        self.d_sino_padded = self.opencl.allocate_array("d_sino_padded", self.sino_padded_shape, dtype=np.float32)
        self.d_sino_f = self.opencl.allocate_array("d_sino_f", self.sino_f_shape, np.complex64)
        self.d_filter_f = self.opencl.allocate_array("d_filter_f", (self.sino_f_shape[-1],), dtype=np.complex64)

    def set_filter(self, h_filt, normalize=True):
        super().set_filter(h_filt, normalize=normalize)
        self.d_filter_f[:] = self.filter_f[:]

    def _init_kernels(self):
        # pointwise complex multiplication
        fname = get_opencl_srcfile("ElementOp.cl")
        if self.ndim == 2:
            kernel_name = "inplace_complex_mul_2Dby1D"
        else:
            kernel_name = "inplace_complex_mul_3Dby1D"
        self.mult_kernel = self.opencl.kernel(kernel_name, filename=fname)
        # padding
        self.padding_kernel = OpenCLPadding(
            self.sino_shape,
            ((0, 0), (self.pad_left, self.pad_right)),
            mode=self.padding_mode,
            opencl_options={"queue": self.queue},
        )
        # memcpy2D
        self.memcpy2D = OpenCLMemcpy2D(queue=self.queue)

    def filter_sino(self, sino, output=None):
        """
        Perform the sinogram siltering.

        Parameters
        ----------
        sino: numpy.ndarray or pyopencl.array
            Input sinogram (2D or 3D)
        output: pyopencl.array, optional
            Output array.
        no_output: bool, optional
            If set to True, no copy is be done. The resulting data lies
            in self.d_sino_padded.
        """
        self._check_array(sino)
        sino = self.opencl.set_array("sino", sino)

        # Padding
        self.padding_kernel.pad(sino, output=self.d_sino_padded)

        # FFT
        self.fft.fft(self.d_sino_padded, self.d_sino_f)

        # multiply padded sinogram with filter in the Fourier domain
        self.mult_kernel(
            self.queue,
            self.d_sino_f,
            self.d_filter_f,
            *(np.int32(self.d_sino_f.shape[::-1])),  # pylint: disable=E1133
            # local_size=None,
            global_size=self.d_sino_f.shape[::-1],
        )  # TODO tune block size ?

        # iFFT
        self.fft.ifft(self.d_sino_f, self.d_sino_padded)

        # return
        if output is None:
            res = self.opencl.allocate_array("output", self.sino_shape)
        else:
            res = output
        if self.ndim == 2:
            # res[:] = self.d_sino_padded[:, self.pad_left : self.pad_left + self.dwidth]
            self.memcpy2D(res, self.d_sino_padded, res.shape[::-1], src_offset_xy=(self.pad_left, 0))
        else:
            res[:] = self.d_sino_padded[:, :, self.pad_left : self.pad_left + self.dwidth]
        return res

    __call__ = filter_sino
